@@ -1,4 +1,4 @@
-# Copyright 2022 The Debusine Developers
+# Copyright 2022-2024 The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -17,7 +17,9 @@ import os
 import re
 import shutil
 import signal
+import subprocess
 import sys
+import tempfile
 from collections.abc import Callable
 from datetime import datetime
 from functools import partial
@@ -25,13 +27,10 @@ from itertools import islice
 from pathlib import Path
 from typing import Any
 
-from dateutil.parser import isoparse
-
-from debian.deb822 import Changes, Dsc
-
-from requests.exceptions import RequestException
-
 import yaml
+from dateutil.parser import isoparse
+from debian.deb822 import Changes, Dsc
+from requests.exceptions import RequestException
 from yaml import YAMLError
 
 from debusine.artifacts import (
@@ -47,7 +46,9 @@ from debusine.client.debusine import Debusine
 from debusine.client.exceptions import DebusineError
 from debusine.client.models import (
     CreateWorkflowRequest,
+    WorkRequestExternalDebsignRequest,
     WorkRequestRequest,
+    WorkRequestResponse,
     WorkflowTemplateRequest,
     model_to_json_serializable_dict,
 )
@@ -68,7 +69,8 @@ class Cli:
         self._argv = argv
 
     @staticmethod
-    def _exit(signum, frame):  # noqa: U100
+    def _exit(signum, frame):
+        signum, frame  # fake usage for vulture
         raise SystemExit(0)
 
     def _parse_args(self):
@@ -87,6 +89,14 @@ class Cli:
             ),
         )
 
+        parser.add_argument(
+            '--scope',
+            help=(
+                'Set scope to be used (use configuration file default '
+                'if not specified)'
+            ),
+        )
+
         parser.add_argument(
             '--config-file',
             default=ConfigHandler.DEFAULT_CONFIG_FILE_PATH,
@@ -168,6 +178,21 @@ class Cli:
             "work_request_id", type=int, help="Work request id to retry"
         )
 
+        provide_signature = subparsers.add_parser(
+            "provide-signature",
+            help="Provide a work request with an external signature",
+        )
+        provide_signature.add_argument(
+            "work_request_id",
+            type=int,
+            help="Work request id that needs a signature",
+        )
+        provide_signature.add_argument(
+            "extra_args",
+            nargs="*",
+            help="Additional arguments passed to debsign",
+        )
+
         create_workflow_template = subparsers.add_parser(
             "create-workflow-template", help="Create a workflow template"
         )
@@ -266,7 +291,7 @@ class Cli:
         )
         import_debian_artifact.add_argument(
             "upload",
-            help=".changes or .dsc to upload",
+            help=".changes, .dsc, or .deb to upload",
         )
 
         download_artifact = subparsers.add_parser(
@@ -296,6 +321,19 @@ class Cli:
             "id", type=int, help="Show information of the artifact"
         )
 
+        show_artifact_relations = subparsers.add_parser(
+            "show-artifact-relations",
+            help="Show artifact relations",
+        )
+        show_artifact_relations.add_argument(
+            "id", type=int, help="Show the relations of an artifact"
+        )
+        show_artifact_relations.add_argument(
+            "--reverse",
+            action="store_true",
+            help="Show reverse relationships, rather than forward ones",
+        )
+
         on_work_request_completed = subparsers.add_parser(
             "on-work-request-completed",
             help="Execute a command when a work request is completed. "
@@ -348,6 +386,7 @@ class Cli:
 
         return Debusine(
             base_api_url=server_configuration['api-url'],
+            scope=self.args.scope or server_configuration['scope'],
             api_token=server_configuration['token'],
             logger=logger,
         )
@@ -408,6 +447,10 @@ class Cli:
                     debusine,
                     self.args.work_request_id,
                 )
+            case "provide-signature":
+                self._provide_signature(
+                    debusine, self.args.work_request_id, self.args.extra_args
+                )
             case "create-artifact":
                 if self.args.data is not None:
                     data = self.args.data.read()
@@ -442,6 +485,10 @@ class Cli:
                 )
             case "show-artifact":
                 self._show_artifact(debusine, self.args.id)
+            case "show-artifact-relations":
+                self._show_artifact_relations(
+                    debusine, self.args.id, self.args.reverse
+                )
             case "on-work-request-completed":
                 self._on_work_request_completed(
                     debusine,
@@ -583,6 +630,93 @@ class Cli:
             work_request_id,
         )
 
+    def _provide_signature_debsign(
+        self,
+        debusine: Debusine,
+        work_request: WorkRequestResponse,
+        debsign_args: list[str],
+    ) -> None:
+        """Provide a work request with an external signature using `debsign`."""
+        # Get a version of the work request with its dynamic task data
+        # resolved.
+        work_request = self._api_call_or_fail(
+            debusine.work_request_external_debsign_get, work_request.id
+        )
+        assert work_request.dynamic_task_data is not None
+        unsigned = self._api_call_or_fail(
+            debusine.artifact_get,
+            work_request.dynamic_task_data["unsigned_id"],
+        )
+
+        with tempfile.TemporaryDirectory(
+            prefix="debusine-debsign-"
+        ) as tmp_name:
+            # Get the individual files that need to be signed.
+            tmp = Path(tmp_name).resolve()
+            for name, file_response in unsigned.files.items():
+                path = tmp / name
+                assert path.resolve().is_relative_to(tmp)
+                assert "/" not in name
+                if (
+                    name.endswith(".changes")
+                    or name.endswith(".dsc")
+                    or name.endswith(".buildinfo")
+                ):
+                    self._api_call_or_fail(
+                        debusine.download_artifact_file, unsigned, name, path
+                    )
+            # Upload artifacts are guaranteed to have exactly one .changes
+            # file.
+            [changes_path] = [
+                path for path in tmp.iterdir() if path.name.endswith(".changes")
+            ]
+
+            # Call debsign.
+            subprocess.run(
+                ["debsign", changes_path.name, *debsign_args],
+                cwd=tmp,
+                check=True,
+            )
+
+            # Create a new artifact and upload the files to it.
+            signed_local = Upload.create(
+                changes_file=changes_path, allow_remote=True
+            )
+            signed_remote = self._api_call_or_fail(
+                debusine.upload_artifact,
+                signed_local,
+                workspace=unsigned.workspace,
+            )
+
+            # Complete the work request using the signed artifact.
+            self._api_call_or_fail(
+                debusine.work_request_external_debsign_complete,
+                work_request.id,
+                WorkRequestExternalDebsignRequest(
+                    signed_artifact=signed_remote.id
+                ),
+            )
+
+    def _provide_signature(
+        self, debusine: Debusine, work_request_id: int, extra_args: list[str]
+    ) -> None:
+        """Provide a work request with an external signature."""
+        # Find out what kind of work request we're dealing with.
+        work_request = self._api_call_or_fail(
+            debusine.work_request_get, work_request_id
+        )
+        match (work_request.task_type, work_request.task_name):
+            case "Wait", "externaldebsign":
+                self._provide_signature_debsign(
+                    debusine, work_request, extra_args
+                )
+            case _:
+                self._fail(
+                    f"Don't know how to provide signature for "
+                    f"{work_request.task_type}/{work_request.task_name} "
+                    f"work request"
+                )
+
     def _create_workflow_template(
         self,
         debusine: Debusine,
@@ -874,6 +1008,9 @@ class Cli:
                 files.append(directory / file["name"])
         return SourcePackage.create(name=name, version=version, files=files)
 
+    def _prepare_deb_for_upload(self, deb: Path) -> BinaryPackage:
+        return BinaryPackage.create(file=deb)
+
     def _import_debian_artifact(
         self,
         debusine: Debusine,
@@ -890,17 +1027,13 @@ class Cli:
                     artifacts = self._prepare_changes_for_upload(path)
                 elif path.suffix == ".dsc":
                     artifacts = [self._prepare_dsc_for_upload(path)]
-                elif path.suffix == ".deb":
-                    raise ValueError(
-                        "Direct uploads of individual .deb files aren't yet "
-                        "possible. However they can be uploaded by using a "
-                        ".changes file."
-                    )
+                elif path.suffix in (".deb", ".ddeb", ".udeb"):
+                    artifacts = [self._prepare_deb_for_upload(path)]
                 else:
                     raise ValueError(
-                        "Only source packages (.dsc) and source/binary uploads "
-                        "(.changes) can be directly imported with this "
-                        f"command. {path} is not supported."
+                        "Only source packages (.dsc), binary packages (.deb), "
+                        "and source/binary uploads (.changes) can be directly "
+                        f"imported with this command. {path} is not supported."
                     )
                 for artifact in artifacts:
                     remote_artifact = self._api_call_or_fail(
@@ -923,8 +1056,8 @@ class Cli:
             for related_artifact in uploaded[1:]:
                 self._api_call_or_fail(
                     debusine.relation_create,
+                    primary_artifact.id,
                     related_artifact.id,
-                    remote_artifact.id,
                     "extends",
                 )
                 result.setdefault("extends", []).append(
@@ -932,8 +1065,8 @@ class Cli:
                 )
                 self._api_call_or_fail(
                     debusine.relation_create,
+                    primary_artifact.id,
                     related_artifact.id,
-                    remote_artifact.id,
                     "relates-to",
                 )
                 result.setdefault("relates_to", []).append(
@@ -973,6 +1106,21 @@ class Cli:
         artifact = self._api_call_or_fail(debusine.artifact_get, artifact_id)
         self._print_yaml(model_to_json_serializable_dict(artifact))
 
+    def _show_artifact_relations(
+        self, debusine: Debusine, artifact_id: int, reverse: bool
+    ) -> None:
+        if reverse:
+            models = self._api_call_or_fail(
+                debusine.relation_list, target_id=artifact_id
+            )
+        else:
+            models = self._api_call_or_fail(
+                debusine.relation_list, artifact_id=artifact_id
+            )
+
+        relations = [model_to_json_serializable_dict(model) for model in models]
+        self._print_yaml(relations)
+
     def _on_work_request_completed(
         self,
         debusine: Debusine,
