@@ -17,12 +17,7 @@ from typing import Any, Optional, TYPE_CHECKING
 from django.conf import settings
 from django.core.exceptions import ValidationError
 from django.db import models, transaction
-from django.db.models import (
-    F,
-    Q,
-    QuerySet,
-    UniqueConstraint,
-)
+from django.db.models import F, Q, QuerySet, UniqueConstraint
 from django.db.models.functions import Coalesce
 from django.urls import reverse
 
@@ -43,6 +38,7 @@ ARTIFACT_CATEGORY_ICON_NAMES = {
     ArtifactCategory.BINARY_PACKAGE: "folder",
     ArtifactCategory.BINARY_PACKAGES: "folder",
     ArtifactCategory.BLHC: "folder",
+    ArtifactCategory.DEBDIFF: "folder",
     ArtifactCategory.LINTIAN: "folder",
     ArtifactCategory.PACKAGE_BUILD_LOG: "journal-text",
     ArtifactCategory.SIGNING_INPUT: "folder",
@@ -60,6 +56,7 @@ ARTIFACT_CATEGORY_SHORT_NAMES = {
     ArtifactCategory.BINARY_PACKAGE: "binary package",
     ArtifactCategory.BINARY_PACKAGES: "binary packages",
     ArtifactCategory.BLHC: "blhc report",
+    ArtifactCategory.DEBDIFF: "debdiff report",
     ArtifactCategory.LINTIAN: "lintian report",
     ArtifactCategory.PACKAGE_BUILD_LOG: "build log",
     ArtifactCategory.SIGNING_INPUT: "signing input",
@@ -98,7 +95,7 @@ class ArtifactManager(models.Manager["Artifact"]):
             file = File.from_local_path(local_path)
             file_store.add_file(local_path, fileobj=file)
             FileInArtifact.objects.create(
-                artifact=artifact, path=artifact_path, file=file
+                artifact=artifact, path=artifact_path, file=file, complete=True
             )
 
         return artifact
@@ -193,6 +190,9 @@ class Artifact(models.Model):
     created_by_work_request = models.ForeignKey(
         "WorkRequest", blank=True, null=True, on_delete=models.SET_NULL
     )
+    original_artifact = models.ForeignKey(
+        "Artifact", blank=True, null=True, on_delete=models.SET_NULL
+    )
 
     objects = ArtifactManager()
 
@@ -283,6 +283,12 @@ class FileInArtifact(models.Model):
     artifact = models.ForeignKey(Artifact, on_delete=models.PROTECT)
     path = models.CharField(max_length=500)
     file = models.ForeignKey(File, on_delete=models.PROTECT)
+    # We do a best-effort data migration here, but note that in cases where
+    # a file store is shared between multiple workspaces, this field may be
+    # left as False for files that were uploaded before this field was
+    # added.  (debusine.debian.net only had a single workspace at the time
+    # of this migration, so that's not a problem there.)
+    complete = models.BooleanField(default=False)
 
     class Meta(TypedModelMeta):
         constraints = [
