@@ -11,24 +11,22 @@
 
 from datetime import datetime
 from functools import cached_property
-from typing import Any, TYPE_CHECKING, TypeAlias
+from typing import Any, ClassVar, Optional, TYPE_CHECKING, TypeAlias
 
+import jsonpath_rw
 from django.conf import settings
 from django.contrib.postgres.constraints import ExclusionConstraint
 from django.contrib.postgres.fields import RangeOperators
 from django.db import models
-from django.db.models import (
-    CheckConstraint,
-    F,
-    Q,
-    QuerySet,
-    UniqueConstraint,
-)
+from django.db.models import CheckConstraint, F, Q, QuerySet, UniqueConstraint
 from django.urls import reverse
 
-import jsonpath_rw
-
-from debusine.artifacts.models import BareDataCategory, CollectionCategory
+from debusine.artifacts.models import (
+    BareDataCategory,
+    CollectionCategory,
+    DebusinePromise,
+    SINGLETON_COLLECTION_CATEGORIES,
+)
 from debusine.db.constraints import JsonDataUniqueConstraint
 from debusine.db.models.workspaces import Workspace
 
@@ -37,6 +35,7 @@ if TYPE_CHECKING:
 
     from debusine.db.models.artifacts import Artifact
     from debusine.db.models.auth import User
+    from debusine.db.models.work_requests import WorkRequest
     from debusine.server.collections import CollectionManagerInterface
 else:
     TypedModelMeta = object
@@ -50,9 +49,31 @@ class _CollectionRetainsArtifacts(models.TextChoices):
     ALWAYS = "always", "Always"
 
 
+class CollectionManager(models.Manager["Collection"]):
+    """Manager for Collection model."""
+
+    @staticmethod
+    def create_singleton(
+        category: CollectionCategory,
+        workspace: Workspace,
+        *,
+        data: dict[str, Any] | None = None,
+    ) -> "Collection":
+        """Create a singleton collection."""
+        if category not in SINGLETON_COLLECTION_CATEGORIES:
+            raise ValueError(
+                f"'{category}' is not a singleton collection category"
+            )
+        return Collection.objects.create(
+            name="_", category=category, workspace=workspace, data=data or {}
+        )
+
+
 class Collection(models.Model):
     """Model representing a collection."""
 
+    objects = CollectionManager()
+
     RetainsArtifacts: TypeAlias = _CollectionRetainsArtifacts
 
     name = models.CharField(max_length=255)
@@ -91,6 +112,16 @@ class Collection(models.Model):
             CheckConstraint(
                 check=~Q(name=""), name="%(app_label)s_%(class)s_name_not_empty"
             ),
+            CheckConstraint(
+                check=(
+                    (
+                        ~Q(category__in=SINGLETON_COLLECTION_CATEGORIES)
+                        & ~Q(name__startswith="_")
+                    )
+                    | Q(category__in=SINGLETON_COLLECTION_CATEGORIES, name="_")
+                ),
+                name="%(app_label)s_%(class)s_name_not_reserved",
+            ),
             CheckConstraint(
                 check=~Q(category=""),
                 name="%(app_label)s_%(class)s_category_not_empty",
@@ -132,8 +163,14 @@ class CollectionItemManager(models.Manager["CollectionItem"]):
         name: str,
         data: dict[str, Any],
         created_by_user: "User",
+        created_by_workflow: Optional["WorkRequest"],
     ) -> "CollectionItem":
         """Create a CollectionItem from bare data."""
+        match category:
+            case BareDataCategory.PROMISE:
+                # Raise ValueError if data is not valid
+                DebusinePromise(**data)
+
         return CollectionItem.objects.create(
             parent_collection=parent_collection,
             name=name,
@@ -141,6 +178,7 @@ class CollectionItemManager(models.Manager["CollectionItem"]):
             category=category,
             data=data,
             created_by_user=created_by_user,
+            created_by_workflow=created_by_workflow,
         )
 
     @staticmethod
@@ -151,6 +189,7 @@ class CollectionItemManager(models.Manager["CollectionItem"]):
         name: str,
         data: dict[str, Any],
         created_by_user: "User",
+        created_by_workflow: Optional["WorkRequest"],
     ) -> "CollectionItem":
         """Create a CollectionItem from the artifact."""
         return CollectionItem.objects.create(
@@ -161,6 +200,7 @@ class CollectionItemManager(models.Manager["CollectionItem"]):
             category=artifact.category,
             data=data,
             created_by_user=created_by_user,
+            created_by_workflow=created_by_workflow,
         )
 
     @staticmethod
@@ -273,8 +313,12 @@ class CollectionItem(models.Model):
         on_delete=models.PROTECT,
         related_name="user_created_%(class)s",
     )
-    # created_by_workflow = models.ForeignKey(Workflow,
-    # on_delete=models.PROTECT, null=True)
+    created_by_workflow = models.ForeignKey(
+        "WorkRequest",
+        on_delete=models.PROTECT,
+        null=True,
+        related_name="workflow_created_%(class)s",
+    )
 
     removed_at = models.DateTimeField(blank=True, null=True)
     removed_by_user = models.ForeignKey(
@@ -283,14 +327,22 @@ class CollectionItem(models.Model):
         null=True,
         related_name="user_removed_%(class)s",
     )
-    # removed_by_workflow = models.ForeignKey(Workflow,
-    # on_delete=models.PROTECT, null=True)
+    removed_by_workflow = models.ForeignKey(
+        "WorkRequest",
+        on_delete=models.PROTECT,
+        null=True,
+        related_name="workflow_removed_%(class)s",
+    )
 
     @staticmethod
     def expand_variables(
-        variables: dict[str, str], artifact: "Artifact"
+        variables: dict[str, str], reference_data: dict[Any, Any]
     ) -> dict[str, str]:
-        """Expand JSONPath variables against an Artifact's data."""
+        """
+        Expand JSONPath variables against some data.
+
+        The data will normally come from an Artifact.
+        """
         jsonpaths = {}
         for name, path in variables.items():
             if name.startswith("$"):
@@ -301,17 +353,17 @@ class CollectionItem(models.Model):
 
         expanded_variables = {}
         for name, jsonpath in jsonpaths.items():
-            matches = jsonpath.find(artifact.data)
+            matches = jsonpath.find(reference_data)
             if len(matches) == 1:
                 expanded_variables[name] = matches[0].value
             elif len(matches) > 1:
                 raise ValueError(
                     "Too many values expanding",
                     variables[f"${name}"],
-                    artifact.data,
+                    reference_data,
                 )
             else:
-                raise KeyError(variables[f"${name}"], artifact.data)
+                raise KeyError(variables[f"${name}"], reference_data)
 
         for name, value in variables.items():
             if not name.startswith("$"):
@@ -440,7 +492,9 @@ class CollectionItemMatchConstraint(models.Model):
     :py:attr:`value`.
     """
 
-    objects = models.Manager["CollectionItemMatchConstraint"]()
+    objects: ClassVar[models.Manager["CollectionItemMatchConstraint"]] = (
+        models.Manager["CollectionItemMatchConstraint"]()
+    )
 
     collection = models.ForeignKey(
         Collection,
