@@ -11,8 +11,8 @@
 
 import json
 import logging
-from collections import defaultdict
-from dataclasses import dataclass
+from collections.abc import Generator
+from contextlib import contextmanager
 from datetime import datetime, timedelta
 from typing import (
     Any,
@@ -22,34 +22,35 @@ from typing import (
     TYPE_CHECKING,
     TypeAlias,
     assert_never,
+    overload,
 )
 
 from django.conf import settings
 from django.core.exceptions import FieldError, ValidationError
+from django.core.serializers.json import DjangoJSONEncoder
 from django.db import models
-from django.db.models import (
-    Count,
-    F,
-    JSONField,
-    Q,
-    QuerySet,
-    UniqueConstraint,
-)
+from django.db.models import F, JSONField, Q, QuerySet, UniqueConstraint
 from django.db.models.functions import Coalesce
 from django.utils import timezone
 
-from debusine.artifacts.models import CollectionCategory
+from debusine.artifacts.models import BareDataCategory, CollectionCategory
+from debusine.client.models import LookupChildType
+from debusine.db.models import CollectionItem
 from debusine.server import notifications
 from debusine.tasks import BaseTask, TaskConfigError
 from debusine.tasks.models import (
+    ActionRetryWithDelays,
     ActionTypes,
     ActionUpdateCollectionWithArtifacts,
     ActionUpdateCollectionWithData,
     EventReaction,
     EventReactions,
+    LookupMultiple,
+    LookupSingle,
     NotificationDataEmail,
     TaskTypes,
 )
+from debusine.tasks.server import TaskDatabaseInterface
 
 if TYPE_CHECKING:
     from django_stubs_ext.db.models import TypedModelMeta
@@ -99,7 +100,7 @@ class WorkflowTemplate(models.Model):
     task_name = models.CharField(
         max_length=100, verbose_name='Name of the Workflow orchestrator class'
     )
-    task_data = JSONField(default=dict, blank=True)
+    task_data = JSONField(default=dict, blank=True, encoder=DjangoJSONEncoder)
     priority = models.IntegerField(
         default=0,
         help_text="Base priority for work requests created from this template",
@@ -142,14 +143,6 @@ class _WorkRequestStatuses(models.TextChoices):
     BLOCKED = "blocked", "Blocked"
 
 
-@dataclass
-class WorkersPendingWorkRequests:
-    """Number of workers and pending work requests."""
-
-    workers: int = 0
-    work_requests_pending: int = 0
-
-
 class WorkRequestManager(models.Manager["WorkRequest"]):
     """Manager for WorkRequest model."""
 
@@ -363,114 +356,6 @@ class WorkRequestManager(models.Manager["WorkRequest"]):
             )
         )
 
-    @classmethod
-    def get_architecture_status(cls) -> dict[str, WorkersPendingWorkRequests]:
-        """
-        Count number of work requests and workers per architecture.
-
-        :return: dictionary with the architectures and number of work requests
-          and workers
-        """
-        result = {}
-
-        workers_per_arch = cls._count_workers_per_architecture()
-
-        # Add architectures with pending work requests
-        for (
-            arch,
-            pending_work_requests,
-        ) in cls._count_pending_work_requests_per_architecture().items():
-            result[arch] = WorkersPendingWorkRequests(
-                workers=workers_per_arch.get(arch, 0),
-                work_requests_pending=pending_work_requests,
-            )
-
-        # Add architectures with online workers, and no pending
-        # work requests (those with pending work requests were already added)
-        listed_archs = result.keys()
-        available_archs = workers_per_arch.keys()
-
-        for arch in available_archs - listed_archs:
-            result[arch] = WorkersPendingWorkRequests(
-                workers=workers_per_arch[arch], work_requests_pending=0
-            )
-
-        return result
-
-    @staticmethod
-    def _count_pending_work_requests_per_architecture() -> dict[str, int]:
-        """Return summary of pending work requests per architecture."""
-        pending_work_requests = (
-            WorkRequest.objects.pending()
-            .filter(task_data__host_architecture__isnull=False)
-            .values("task_data__host_architecture")
-            .annotate(count=Count("id"))
-            .order_by()
-        )
-
-        result: defaultdict[str, int] = defaultdict(int)
-
-        # Add architectures with pending work requests if their
-        # "task_data" contains "host_architecture"
-
-        # TODO: count work requests per architecture using WorkRequest tags
-        # (when tags are implemented in Debusine) instead of using
-        # task_data__host_architecture or instantiating Tasks and calling
-        # host_architecture() method
-
-        for pending_host_architecture in pending_work_requests:
-            result[
-                pending_host_architecture["task_data__host_architecture"]
-            ] = pending_host_architecture["count"]
-
-        # Add architectures with pending work requests if their
-        # "task_data" does not contain "host_architecture".
-        pending_work_requests_no_host_architecture = (
-            WorkRequest.objects.pending()
-            .filter(task_data__host_architecture__isnull=True)
-            .order_by()
-        )
-
-        for (
-            pending_no_host_architecture
-        ) in pending_work_requests_no_host_architecture:
-            try:
-                task = pending_no_host_architecture.get_task()
-            except TaskConfigError:
-                # When the scheduler picks this work request: will mark it
-                # as aborted
-                continue
-
-            host_architecture = task.host_architecture()
-
-            if host_architecture is None:
-                continue
-
-            result[host_architecture] += 1
-
-        return dict(result)
-
-    @staticmethod
-    def _count_workers_per_architecture() -> dict[str, int]:
-        """
-        Return number of connected workers per architecture.
-
-        The same worker might be able to execute tasks for multiple
-        architectures.
-        """
-        from debusine.db.models.workers import Worker
-
-        result: defaultdict[str, int] = defaultdict(int)
-
-        for worker in Worker.objects.filter(token__enabled=True):
-            for arch in worker.metadata().get("system:architectures", []):
-                if worker.connected():
-                    result[arch] += 1
-                else:
-                    result.setdefault(arch, 0)
-
-        return dict(result)
-
 
 class WorkRequest(models.Model):
     """
@@ -570,8 +455,10 @@ class WorkRequest(models.Model):
     task_name = models.CharField(
         max_length=100, verbose_name='Name of the task to execute'
     )
-    task_data = JSONField(default=dict, blank=True)
-    dynamic_task_data = JSONField(null=True, blank=True)
+    task_data = JSONField(default=dict, blank=True, encoder=DjangoJSONEncoder)
+    dynamic_task_data = JSONField(
+        null=True, blank=True, encoder=DjangoJSONEncoder
+    )
     priority_base = models.IntegerField(
         default=0, help_text="Base priority of this work request"
     )
@@ -603,10 +490,16 @@ class WorkRequest(models.Model):
         "self", symmetrical=False, related_name="reverse_dependencies"
     )
     workflow_data_json = JSONField(
-        default=dict, blank=True, db_column="workflow_data"
+        default=dict,
+        blank=True,
+        db_column="workflow_data",
+        encoder=DjangoJSONEncoder,
     )
     event_reactions_json = JSONField(
-        default=dict, blank=True, db_column="event_reactions"
+        default=dict,
+        blank=True,
+        db_column="event_reactions",
+        encoder=DjangoJSONEncoder,
     )
     internal_collection = models.OneToOneField(
         "Collection",
@@ -698,6 +591,16 @@ class WorkRequest(models.Model):
         """Set event_reactions_json from a pydantic structure."""
         self.event_reactions_json = value.dict()
 
+    @contextmanager
+    def scheduling_disabled(self) -> Generator[None, None, None]:
+        """Temporarily disable scheduling on changes to this work request."""
+        try:
+            # See debusine.server.scheduler._work_request_changed.
+            self._disable_signals = True
+            yield
+        finally:
+            del self._disable_signals
+
     def create_child(
         self,
         task_name: str,
@@ -728,8 +631,6 @@ class WorkRequest(models.Model):
 
     def can_satisfy_dynamic_data(self) -> bool:
         """Check if dynamic data can still be looked up."""
-        from debusine.server.scheduler import TaskDatabase
-
         try:
             task = self.get_task()
         except TaskConfigError:
@@ -787,16 +688,27 @@ class WorkRequest(models.Model):
 
     def _retry_workflow(self) -> Self:
         """Retry logic for workflows."""
-        from debusine.server.scheduler import populate_workflow
+        # Import here to prevent circular imports
+        from debusine.server.workflows.base import orchestrate_workflow
+        from debusine.server.workflows.models import WorkRequestWorkflowData
 
         # Set the workflow back to running, so that further actions execute in
         # a running workflow
         self.status = self.Statuses.PENDING
         self.mark_running()
 
+        # Increment the retry count
+        if (workflow_data := self.workflow_data) is None:
+            workflow_data = WorkRequestWorkflowData()
+        workflow_data.retry_count += 1
+        self.workflow_data = workflow_data
+        self.save()
+
         try:
+            # TODO: workflow orchestrators may be slow; this should move to
+            # a Celery task
             # Give the workflow a chance to update its graph
-            populate_workflow(self)
+            orchestrate_workflow(self)
 
             # Retry failed child work requests
             for child in (
@@ -817,6 +729,9 @@ class WorkRequest(models.Model):
 
     def _retry_supersede(self) -> "WorkRequest":
         """Retry logic superseding work requests."""
+        # Import here to prevent circular imports
+        from debusine.server.workflows.models import WorkRequestWorkflowData
+
         work_request = WorkRequest.objects.create(
             workspace=self.workspace,
             created_by=self.created_by,
@@ -853,9 +768,31 @@ class WorkRequest(models.Model):
             workflow.completed_at = None
             workflow.save()
 
+        # Increment the retry count
+        if (workflow_data := work_request.workflow_data) is None:
+            workflow_data = WorkRequestWorkflowData()
+        workflow_data.retry_count += 1
+        work_request.workflow_data = workflow_data
+        work_request.save()
+
         if work_request.can_be_automatically_unblocked():
             work_request.mark_pending()
 
+        # Update promises associated with the previous work request
+        if (
+            self.parent is not None
+            and self.parent.internal_collection is not None
+        ):
+            for debusine_promise in CollectionItem.objects.filter(
+                category=BareDataCategory.PROMISE,
+                parent_collection=self.parent.internal_collection,
+                data__promise_work_request_id=self.id,
+            ):
+                debusine_promise.data["promise_work_request_id"] = (
+                    work_request.id
+                )
+                debusine_promise.save()
+
         return work_request
 
     def retry(self) -> "WorkRequest":
@@ -879,7 +816,12 @@ class WorkRequest(models.Model):
             )
 
         match self.task_type:
-            case TaskTypes.WORKER | TaskTypes.SERVER | TaskTypes.SIGNING:
+            case (
+                TaskTypes.WORKER
+                | TaskTypes.SERVER
+                | TaskTypes.SIGNING
+                | TaskTypes.WAIT
+            ):
                 try:
                     task_cls = BaseTask.class_from_name(
                         self.task_type, self.task_name
@@ -941,33 +883,60 @@ class WorkRequest(models.Model):
                     f"task type {self.task_type} not yet supported"
                 )
 
-    def get_task(self) -> BaseTask[Any, Any]:
+    def get_task(self, worker: "Worker | None" = None) -> BaseTask[Any, Any]:
         """
         Instantiate the Task for this work request.
 
-        :raise InternalTaskError: if the work request is for an internal task
+        :param worker: if set, the worker that this work request is intended
+          to run on; `worker_host_architecture` will be set to its host
+          architecture
+        :raise InternalTaskError: if the work request is for an internal
+          task other than a workflow callback
         """
-        match self.task_type:
-            case TaskTypes.WORKER | TaskTypes.SERVER | TaskTypes.SIGNING:
+        # Import here to prevent circular imports
+        from debusine.server.workflows import Workflow
+
+        match (self.task_type, self.task_name):
+            case (
+                (TaskTypes.WORKER, _)
+                | (TaskTypes.SERVER, _)
+                | (TaskTypes.SIGNING, _)
+                | (TaskTypes.WAIT, _)
+            ):
                 task_cls = BaseTask.class_from_name(
-                    self.task_type, self.task_name
+                    TaskTypes(self.task_type), self.task_name
                 )
-                return task_cls(
+                task = task_cls(
                     task_data=self.task_data,
                     dynamic_task_data=self.dynamic_task_data,
                 )
-            case TaskTypes.WORKFLOW:
-                # Import here to prevent circular imports
-                from debusine.server.workflows import Workflow
-
+            case (TaskTypes.WORKFLOW, _):
                 workflow_cls = Workflow.from_name(self.task_name)
-                return workflow_cls(self)
-            case TaskTypes.INTERNAL:
-                raise InternalTaskError("Internal tasks cannot be instantiated")
-            case _ as unreachable:
-                raise NotImplementedError(
-                    f"task type {unreachable} not yet supported"
+                task = workflow_cls(self)
+            case (TaskTypes.INTERNAL, "workflow"):
+                if (workflow := self.parent) is None:
+                    raise InternalTaskError(
+                        "Workflow callback is not contained in a workflow"
+                    )
+                assert workflow.task_type == TaskTypes.WORKFLOW
+                workflow_cls = Workflow.from_name(workflow.task_name)
+                task = workflow_cls(workflow)
+            case (TaskTypes.INTERNAL, _):
+                raise InternalTaskError(
+                    "Internal tasks other than workflow callbacks cannot be "
+                    "instantiated"
                 )
+            case _ as unreachable:  # noqa: F841
+                raise NotImplementedError(
+                    f"task type {self.task_type} not yet supported"
+                )
+
+        if worker is not None:
+            task.worker_host_architecture = worker.metadata().get(
+                "system:host_architecture"
+            )
+
+        return task
 
     def mark_running(self) -> bool:
         """Worker has begun executing the task."""
@@ -1067,6 +1036,7 @@ class WorkRequest(models.Model):
         self.process_update_collection_with_data(
             actions[ActionTypes.UPDATE_COLLECTION_WITH_DATA]
         )
+        self.process_retry_with_delays(actions[ActionTypes.RETRY_WITH_DELAYS])
 
     def get_triggered_actions(
         self,
@@ -1108,7 +1078,7 @@ class WorkRequest(models.Model):
                     self.workspace,
                     user=self.created_by,
                     workflow_root=self.get_workflow_root(),
-                    expect_type=CollectionItem.Types.COLLECTION,
+                    expect_type=LookupChildType.COLLECTION,
                 ).collection
             except LookupError as e:
                 logger.error("%s in WorkRequest %s", e, self.pk)  # noqa: G200
@@ -1131,7 +1101,7 @@ class WorkRequest(models.Model):
             for artifact in artifacts_to_add:
                 try:
                     expanded_variables = CollectionItem.expand_variables(
-                        update.variables or {}, artifact
+                        update.variables or {}, artifact.data
                     )
                 except (KeyError, ValueError):
                     logger.exception(
@@ -1154,6 +1124,7 @@ class WorkRequest(models.Model):
                     manager.add_artifact(
                         artifact,
                         user=self.created_by,
+                        workflow=self.parent,
                         variables=item_variables,
                         name=item_name,
                         replace=True,
@@ -1188,7 +1159,7 @@ class WorkRequest(models.Model):
                     self.workspace,
                     user=self.created_by,
                     workflow_root=self.get_workflow_root(),
-                    expect_type=CollectionItem.Types.COLLECTION,
+                    expect_type=LookupChildType.COLLECTION,
                 ).collection
             except LookupError as e:
                 logger.error("%s in WorkRequest %s", e, self.pk)  # noqa: G200
@@ -1208,6 +1179,7 @@ class WorkRequest(models.Model):
                 manager.add_bare_data(
                     update.category,
                     user=self.created_by,
+                    workflow=self.parent,
                     data=data,
                     name=item_name,
                     replace=True,
@@ -1221,6 +1193,51 @@ class WorkRequest(models.Model):
                     self.pk,
                 )
 
+    def process_retry_with_delays(self, actions: list[EventReaction]) -> None:
+        """Retry a work request with delays."""
+        # Import here to prevent circular imports
+        from debusine.server.tasks.wait.models import DelayData
+
+        for action in actions:
+            assert isinstance(action, ActionRetryWithDelays)
+            if self.parent is None:
+                raise CannotRetry(
+                    "retry-with-delays action may only be used in a workflow"
+                )
+            retry_count = (
+                0
+                if self.workflow_data is None
+                else self.workflow_data.retry_count
+            )
+            if retry_count >= len(action.delays):
+                continue
+            m = ActionRetryWithDelays._delay_re.match(
+                action.delays[retry_count]
+            )
+            # Checked by ActionRetryWithDelays.validate_delays.
+            assert m is not None
+            match m.group(2):
+                case "m":
+                    delay = timedelta(minutes=int(m.group(1)))
+                case "h":
+                    delay = timedelta(hours=int(m.group(1)))
+                case "d":
+                    delay = timedelta(days=int(m.group(1)))
+                case "w":
+                    delay = timedelta(weeks=int(m.group(1)))
+                case _ as unreachable:
+                    raise AssertionError(
+                        f"Unexpected delay unit: {unreachable}"
+                    )
+            wr_delay = self.parent.create_child(
+                task_type=TaskTypes.WAIT,
+                task_name="delay",
+                task_data=DelayData(delay_until=timezone.now() + delay).dict(),
+                status=self.Statuses.PENDING,
+            )
+            wr_retried = self.retry()
+            wr_retried.add_dependency(wr_delay)
+
     def mark_pending(self) -> bool:
         """Worker is ready for execution."""
         if self.status != self.Statuses.BLOCKED:
@@ -1405,6 +1422,17 @@ class WorkRequest(models.Model):
                 if not self.mark_aborted():
                     raise AssertionError("Work request could not be aborted")
 
+    @property
+    def requires_signature(self) -> bool:
+        """True if this task is blocked on `debusine provide-signature`."""
+        # TODO: At some point it may be worth having this delegate to a
+        # property on the task, but for now this is good enough.
+        return (
+            self.task_type == TaskTypes.WAIT
+            and self.task_name == "externaldebsign"
+            and self.status == self.Statuses.BLOCKED
+        )
+
     def mark_aborted(self) -> bool:
         """
         Worker has aborted the task after request from UI.
@@ -1414,6 +1442,7 @@ class WorkRequest(models.Model):
         self.completed_at = timezone.now()
         self.status = self.Statuses.ABORTED
         self.save()
+        self.unblock_reverse_dependencies()
         if self.parent is not None:
             self.parent.maybe_finish_workflow()
         logger.debug(
@@ -1475,6 +1504,11 @@ class WorkRequest(models.Model):
         """Return whether this work request is part of a workflow."""
         return self.parent is not None
 
+    @property
+    def is_workflow_root(self) -> bool:
+        """Return whether this work request is a root workflow."""
+        return self.is_workflow and self.parent is None
+
     def get_workflow_root(self) -> Optional["WorkRequest"]:
         """Return the root of this work request's workflow, if any."""
         if not self.is_workflow and not self.is_part_of_workflow:
@@ -1487,7 +1521,10 @@ class WorkRequest(models.Model):
     @property
     def workflow_display_name(self) -> str:
         """Return this work request's name for display in a workflow."""
-        if self.workflow_data is not None:
+        if (
+            self.workflow_data is not None
+            and self.workflow_data.display_name is not None
+        ):
             return self.workflow_data.display_name
         else:
             return self.task_name
@@ -1538,6 +1575,117 @@ class WorkRequest(models.Model):
         return task.get_label()
 
 
+class TaskDatabase(TaskDatabaseInterface):
+    """Implementation of database interaction in worker tasks."""
+
+    def __init__(self, work_request: WorkRequest) -> None:
+        """Construct a :py:class:`TaskDatabase`."""
+        self.work_request = work_request
+
+    @overload
+    def lookup_single_artifact(
+        self,
+        lookup: LookupSingle,
+        default_category: CollectionCategory | None = None,
+    ) -> int: ...
+
+    @overload
+    def lookup_single_artifact(
+        self,
+        lookup: None,
+        default_category: CollectionCategory | None = None,
+    ) -> None: ...
+
+    def lookup_single_artifact(
+        self,
+        lookup: LookupSingle | None,
+        default_category: CollectionCategory | None = None,
+    ) -> int | None:
+        """Look up a single artifact using :ref:`lookup-single`."""
+        # Import here to prevent circular imports
+        from debusine.server.collections.lookup import lookup_single
+
+        return (
+            None
+            if lookup is None
+            else lookup_single(
+                lookup=lookup,
+                workspace=self.work_request.workspace,
+                user=self.work_request.created_by,
+                default_category=default_category,
+                workflow_root=self.work_request.get_workflow_root(),
+                expect_type=LookupChildType.ARTIFACT,
+            ).artifact.id
+        )
+
+    def lookup_multiple_artifacts(
+        self,
+        lookup: LookupMultiple | None,
+        default_category: CollectionCategory | None = None,
+    ) -> list[int]:
+        """Look up multiple artifacts using :ref:`lookup-multiple`."""
+        # Import here to prevent circular imports
+        from debusine.server.collections.lookup import lookup_multiple
+
+        return (
+            []
+            if lookup is None
+            else [
+                result.artifact.id
+                for result in lookup_multiple(
+                    lookup=lookup,
+                    workspace=self.work_request.workspace,
+                    user=self.work_request.created_by,
+                    default_category=default_category,
+                    workflow_root=self.work_request.get_workflow_root(),
+                    expect_type=LookupChildType.ARTIFACT,
+                )
+            ]
+        )
+
+    @overload
+    def lookup_single_collection(
+        self,
+        lookup: LookupSingle,
+        default_category: CollectionCategory | None = None,
+    ) -> int: ...
+
+    @overload
+    def lookup_single_collection(
+        self,
+        lookup: None,
+        default_category: CollectionCategory | None = None,
+    ) -> None: ...
+
+    def lookup_single_collection(
+        self,
+        lookup: LookupSingle | None,
+        default_category: CollectionCategory | None = None,
+    ) -> int | None:
+        """Look up a single collection using :ref:`lookup-single`."""
+        # Import here to prevent circular imports
+        from debusine.server.collections.lookup import lookup_single
+
+        return (
+            None
+            if lookup is None
+            else lookup_single(
+                lookup=lookup,
+                workspace=self.work_request.workspace,
+                user=self.work_request.created_by,
+                default_category=default_category,
+                workflow_root=self.work_request.get_workflow_root(),
+                expect_type=LookupChildType.COLLECTION,
+            ).collection.id
+        )
+
+    def get_server_setting(self, setting: str) -> str:
+        """Look up a Django setting (strings only)."""
+        value = getattr(settings, setting)
+        assert isinstance(value, str)
+        return value
+
+
 class NotificationChannel(models.Model):
     """Model to store notification configuration."""
 
