@@ -11,12 +11,16 @@
 
 import tempfile
 from datetime import timedelta
+from functools import cached_property
 from pathlib import Path
-from typing import Any
+from typing import Any, overload
 
+from django.conf import settings
 from django.contrib.auth.models import Permission
 from django.contrib.contenttypes.models import ContentType
 from django.db.models import Model
+from django.test import override_settings
+from django.utils import timezone
 
 from debusine.artifacts.models import (
     ArtifactCategory,
@@ -37,6 +41,9 @@ from debusine.db.models import (
     FileInArtifact,
     FileStore,
     FileUpload,
+    Group,
+    Scope,
+    TaskDatabase,
     Token,
     User,
     WorkRequest,
@@ -44,10 +51,12 @@ from debusine.db.models import (
     WorkflowTemplate,
     Workspace,
 )
+from debusine.db.models.permissions import Roles
+from debusine.db.models.scopes import ScopeRoles
+from debusine.db.models.workspaces import WorkspaceRoles
 from debusine.db.tests.utils import _calculate_hash_from_data
 from debusine.server.collections.lookup import lookup_single
 from debusine.server.file_backend.interface import FileBackendInterface
-from debusine.server.scheduler import TaskDatabase
 from debusine.server.workflows.models import WorkRequestWorkflowData
 from debusine.tasks.models import (
     BackendType,
@@ -102,6 +111,15 @@ class Playground(ArtifactPlayground):
                 configuration={},
             )
 
+    @cached_property
+    def default_scope(self) -> Scope:
+        """Get the default workspace used by create methods."""
+        # TODO: when tests pass with scoped workspaces, switch to a differently
+        # named default scope for tests, to catch code possibly depending on
+        # the fallback scope introduced by the migration
+        # return self.get_or_create_scope(name="tests")
+        return self.get_or_create_scope(name=settings.DEBUSINE_DEFAULT_SCOPE)
+
     def get_default_workspace(self) -> Workspace:
         """Get the default workspace used by create methods."""
         return self.create_workspace()
@@ -117,6 +135,70 @@ class Playground(ArtifactPlayground):
             )
         )
 
+    def get_or_create_scope(self, name: str) -> Scope:
+        """Create a scope, or return an existing scope."""
+        return Scope.objects.get_or_create(name=name)[0]
+
+    def create_user(
+        self,
+        username: str,
+    ) -> User:
+        """Create a user."""
+        email = f"{username}@example.org"
+        user = User.objects.create_user(
+            username=username,
+            email=email,
+        )
+        return user
+
+    @overload
+    def create_group_role(
+        self,
+        resource: Scope,
+        role: ScopeRoles,
+        *users: User,
+        name: str | None = None,
+    ) -> Group: ...
+
+    @overload
+    def create_group_role(
+        self,
+        resource: Workspace,
+        role: WorkspaceRoles,
+        *users: User,
+        name: str | None = None,
+    ) -> Group: ...
+
+    def create_group_role(
+        self,
+        resource: Model,
+        role: Roles,
+        *users: User,
+        name: str | None = None,
+    ) -> Group:
+        """Create a group with the given role on a resource."""
+        if isinstance(resource, Scope):
+            group_scope = resource
+        elif hasattr(resource, "scope"):
+            group_scope = resource.scope
+        # Currently not used
+        # elif hasattr(resource, "workspace"):
+        #     group_scope = resource.workspace.scope
+        else:
+            raise NotImplementedError(
+                "unable to get scope for"
+                f" {resource.__class__.__name__} resources"
+            )
+
+        if name is None:
+            name = f"{resource} {role}"
+
+        group = Group.objects.create(scope=group_scope, name=name)
+        getattr(resource, "assign_role")(role, group)
+        for user in users:
+            group.users.add(user)
+        return group
+
     def create_token_enabled(
         self, with_user: bool = False, **kwargs: Any
     ) -> Token:
@@ -136,13 +218,20 @@ class Playground(ArtifactPlayground):
         return token
 
     def create_workspace(
-        self, *, file_store: FileStore | None = None, **kwargs: Any
+        self,
+        *,
+        file_store: FileStore | None = None,
+        scope: Scope | None = None,
+        **kwargs: Any,
     ) -> Workspace:
         """Create a Workspace and return it."""
         if file_store is None:
             file_store = self.get_default_file_store()
+        if scope is None:
+            scope = self.default_scope
         workspace_name = kwargs.pop("name", self.default_workspace_name)
         workspace, _ = Workspace.objects.get_or_create(
+            scope=scope,
             name=workspace_name,
             defaults={**{"default_file_store": file_store, **kwargs}},
         )
@@ -165,6 +254,20 @@ class Playground(ArtifactPlayground):
             data=data or {},
         )
 
+    def create_singleton_collection(
+        self,
+        category: CollectionCategory,
+        *,
+        workspace: Workspace | None = None,
+        data: dict[str, Any] | None = None,
+    ) -> Collection:
+        """Create a singleton collection."""
+        return Collection.objects.create_singleton(
+            category=category,
+            workspace=workspace or self.get_default_workspace(),
+            data=data or {},
+        )
+
     def create_workflow_template(
         self,
         name: str,
@@ -183,11 +286,37 @@ class Playground(ArtifactPlayground):
             **kwargs,
         )
 
+    def create_workflow(
+        self,
+        task_name: str = "noop",
+        task_data: dict[str, Any] | None = None,
+        parent: WorkRequest | None = None,
+        status: WorkRequest.Statuses | None = None,
+    ) -> WorkRequest:
+        """Create a workflow."""
+        workflow_template_name = f"{task_name}-template"
+        try:
+            template = WorkflowTemplate.objects.get(name=workflow_template_name)
+        except WorkflowTemplate.DoesNotExist:
+            template = self.create_workflow_template(
+                name=workflow_template_name, task_name=task_name
+            )
+
+        return WorkRequest.objects.create_workflow(
+            template=template,
+            created_by=self.get_default_user(),
+            data=task_data or {},
+            parent=parent,
+            status=status,
+        )
+
     def create_work_request(
         self,
         mark_running: bool = False,
         assign_new_worker: bool = False,
         result: WorkRequest.Results | None = None,
+        expired: bool = False,
+        workspace: Workspace | None = None,
         **kwargs: Any,
     ) -> WorkRequest:
         """
@@ -196,6 +325,8 @@ class Playground(ArtifactPlayground):
         :param mark_running: if True call mark_running() method
         :param assign_new_worker: if not None assign worker to the work request
         :param result: if not None call mark_completed(result)
+        :param expired: True to create an expired work request
+          (created_at: 1 year ago, expiration_delay: 1 day)
         :param kwargs: use them when creating the WorkRequest model
         """
         # FileStore and Workspace are usually created by the migrations
@@ -204,16 +335,22 @@ class Playground(ArtifactPlayground):
         # yet (need verification, but they seem to be in a non-committed
         # transaction and the test code is in a different thread because
         # of the implementation of database_sync_to_async
-        workspace = self.create_workspace()
+        if workspace is None:
+            workspace = self.create_workspace()
 
         defaults: dict[str, Any] = {"workspace": workspace}
         if "created_by" not in kwargs:
             token = self.create_token_enabled(with_user=True)
             defaults["created_by"] = token.user
+        if expired:
+            defaults.update(
+                created_at=timezone.now() - timedelta(days=365),
+                expiration_delay=timedelta(days=1),
+            )
 
         defaults.update(kwargs)
 
-        created_at = kwargs.pop("created_at", None)
+        created_at = defaults.pop("created_at", None)
 
         work_request = WorkRequest.objects.create(**defaults)
 
@@ -393,7 +530,10 @@ class Playground(ArtifactPlayground):
                     )
 
                 FileInArtifact.objects.create(
-                    artifact=artifact, path=path, file=fileobj
+                    artifact=artifact,
+                    path=path,
+                    file=fileobj,
+                    complete=not skip_add_files_in_store,
                 )
 
         return artifact, files_contents
@@ -458,6 +598,26 @@ class Playground(ArtifactPlayground):
             )
             return source
 
+    def create_minimal_binary_packages_artifact(
+        self,
+        srcpkg_name: str,
+        srcpkg_version: str,
+        version: str,
+        architecture: str,
+    ) -> Artifact:
+        """Create a minimal `debian:binary-packages` artifact."""
+        artifact, _ = self.create_artifact(
+            category=ArtifactCategory.BINARY_PACKAGES,
+            data={
+                "srcpkg_name": srcpkg_name,
+                "srcpkg_version": srcpkg_version,
+                "version": version,
+                "architecture": architecture,
+                "packages": [],
+            },
+        )
+        return artifact
+
     def create_build_log_artifact(
         self,
         *,
@@ -468,6 +628,7 @@ class Playground(ArtifactPlayground):
         work_request: WorkRequest | None = None,
         created_by: User | None = None,
         contents: bytes | None = None,
+        skip_add_files_in_store: bool = False,
     ) -> Artifact:
         """Create an artifact for a build log."""
         filename = f"{source}_{version}_{build_arch}.buildlog"
@@ -491,6 +652,7 @@ class Playground(ArtifactPlayground):
             work_request=work_request,
             created_by=created_by,
             create_files=True,
+            skip_add_files_in_store=skip_add_files_in_store,
         )
         return artifact
 
@@ -615,6 +777,7 @@ class Playground(ArtifactPlayground):
                     else SbuildBuildComponent.ANY
                 )
             ],
+            build_profiles=None,
         )
         if workflow:
             return workflow.create_child(
@@ -632,11 +795,17 @@ class Playground(ArtifactPlayground):
                 workspace=workspace or self.get_default_workspace(),
             )
 
-    def compute_dynamic_data(self, work_request: WorkRequest) -> None:
+    def compute_dynamic_data(
+        self, work_request: WorkRequest, worker: Worker | None = None
+    ) -> None:
         """Compute dynamic data for the work request, and save it."""
-        task = work_request.get_task()
-        if data := task.compute_dynamic_data(TaskDatabase(work_request)):
-            work_request.dynamic_task_data = data.dict()
+        task = work_request.get_task(worker=worker)
+        with override_settings(DEBUSINE_FQDN="debusine-playground.net"):
+            data = task.compute_dynamic_data(TaskDatabase(work_request))
+        if data:
+            work_request.dynamic_task_data = model_to_json_serializable_dict(
+                data
+            )
         else:
             work_request.dynamic_task_data = None
         work_request.save()
@@ -670,13 +839,13 @@ class Playground(ArtifactPlayground):
             workflow=workflow,
             workspace=workspace,
         )
-
-        # Resolve dynamic data
-        self.compute_dynamic_data(work_request)
-
-        # Assign a worker
         if worker is None:
             worker = self.create_worker()
+
+        # Resolve dynamic data
+        self.compute_dynamic_data(work_request, worker)
+
+        # Assign a worker
         work_request.assign_worker(worker)
         work_request.mark_running()
 
