@@ -11,13 +11,16 @@
 
 import django.test
 from django.contrib.auth.hashers import check_password
+from django.utils import timezone
 
 from debusine.artifacts.models import (
     ArtifactCategory,
     CollectionCategory,
+    DebianBinaryPackages,
     DebianPackageBuildLog,
     DebianSourcePackage,
 )
+from debusine.db.context import context
 from debusine.db.models import (
     Artifact,
     ArtifactRelation,
@@ -91,6 +94,20 @@ class PlaygroundTest(django.test.TestCase, TestCase):
         self.assertTrue(user.has_usable_password())
         self.assertTrue(check_password("test", user.password))
 
+    @context.disable_permission_checks()
+    def test_create_workspace_scope(self) -> None:
+        """Test setting scope on created workspaces."""
+        playground = Playground()
+        testscope = playground.get_or_create_scope("testscope")
+
+        ws1 = playground.create_workspace()
+        self.assertEqual(playground.create_workspace(), ws1)
+
+        ws2 = playground.create_workspace(scope=testscope)
+        self.assertEqual(playground.create_workspace(scope=testscope), ws2)
+        self.assertNotEqual(ws1, ws2)
+
+    @context.disable_permission_checks()
     def test_create_workspace_file_store(self) -> None:
         """Test setting file store on created workspaces."""
         playground = Playground()
@@ -141,6 +158,9 @@ class PlaygroundTest(django.test.TestCase, TestCase):
                 "environment_id": environment.pk,
                 "input_extra_binary_artifacts_ids": [],
                 "input_source_artifact_id": source.pk,
+                "binnmu_maintainer": (
+                    "Debusine <noreply@debusine-playground.net>"
+                ),
             },
         )
 
@@ -157,6 +177,25 @@ class PlaygroundTest(django.test.TestCase, TestCase):
         artifact = DebianSourcePackage(**source.data)
         self.assert_source_artifact_equal(artifact, "hello", "1.0-1")
 
+    def test_create_minimal_binary_packages_artifact(self) -> None:
+        """Test creating a binary_packages artifact."""
+        playground = Playground()
+        bp = playground.create_minimal_binary_packages_artifact(
+            "hello", "1.0-1", "1.0-1", "amd64"
+        )
+        self.assertEqual(bp.category, ArtifactCategory.BINARY_PACKAGES)
+        self.assertEqual(bp.workspace, playground.get_default_workspace())
+        self.assertEqual(bp.files.count(), 0)
+        self.assertEqual(bp.created_by, None)
+        self.assertIsNone(bp.created_by_work_request)
+
+        artifact = DebianBinaryPackages(**bp.data)
+        self.assertEqual(artifact.srcpkg_name, "hello")
+        self.assertEqual(artifact.srcpkg_version, "1.0-1")
+        self.assertEqual(artifact.version, "1.0-1")
+        self.assertEqual(artifact.architecture, "amd64")
+        self.assertEqual(artifact.packages, [])
+
     def test_create_source_artifact_with_files(self) -> None:
         """Test creating a source artifact with its files."""
         playground = Playground()
@@ -190,7 +229,7 @@ class PlaygroundTest(django.test.TestCase, TestCase):
         backend = buildlog.workspace.default_file_store.get_backend_object()
         with backend.get_stream(file) as fd:
             self.assertEqual(
-                fd.readlines()[3].decode(),
+                fd.read().splitlines(keepends=True)[3].decode(),
                 "Line 4 of hello_1.0-1_amd64.buildlog\n",
             )
 
@@ -248,6 +287,7 @@ class PlaygroundTest(django.test.TestCase, TestCase):
         self.assertEqual(env.category, CollectionCategory.ENVIRONMENTS)
         self.assertEqual(env.workspace, playground.get_default_workspace())
 
+    @context.disable_permission_checks()
     def test_create_debian_env_collection_custom(self) -> None:
         """Test create_debian_environments_collection."""
         playground = Playground()
@@ -301,6 +341,7 @@ class PlaygroundTest(django.test.TestCase, TestCase):
             },
         )
 
+    @context.disable_permission_checks()
     def test_create_debian_env_custom(self) -> None:
         """Test create_debian_environment with custom args."""
         playground = Playground()
@@ -424,6 +465,7 @@ class PlaygroundTest(django.test.TestCase, TestCase):
                 'backend': BackendType.UNSHARE,
                 'binnmu': None,
                 'build_components': [SbuildBuildComponent.ALL],
+                'build_profiles': None,
                 'distribution': None,
                 'environment': environment_item.artifact.pk,
                 'host_architecture': 'all',
@@ -477,6 +519,7 @@ class PlaygroundTest(django.test.TestCase, TestCase):
                 'backend': BackendType.UNSHARE,
                 'binnmu': None,
                 'build_components': [SbuildBuildComponent.ALL],
+                'build_profiles': None,
                 'distribution': None,
                 'environment': environment.pk,
                 'host_architecture': 'all',
@@ -487,10 +530,11 @@ class PlaygroundTest(django.test.TestCase, TestCase):
             },
         )
         self.assertEqual(
-            wr.dynamic_task_data,
+            SbuildDynamicData.parse_obj(wr.dynamic_task_data),
             SbuildDynamicData(
                 environment_id=environment.pk,
                 input_source_artifact_id=source.pk,
+                binnmu_maintainer="Debusine <noreply@debusine-playground.net>",
             ),
         )
         self.assertIsNone(wr.parent)
@@ -549,11 +593,13 @@ class PlaygroundTest(django.test.TestCase, TestCase):
             binary, [(source, ArtifactRelation.Relations.BUILT_USING)]
         )
 
+    @context.disable_permission_checks()
     def test_simulate_package_build_custom(self) -> None:
         """Test simulating a whole package build with custom args."""
         playground = Playground()
 
-        workspace = playground.create_workspace(name="custom")
+        with context.disable_permission_checks():
+            workspace = playground.create_workspace(name="custom", public=True)
         source = playground.create_source_artifact(
             name="test", version="2.0", workspace=workspace
         )
@@ -608,6 +654,7 @@ class PlaygroundTest(django.test.TestCase, TestCase):
                 'backend': BackendType.UNSHARE,
                 'binnmu': None,
                 'build_components': [SbuildBuildComponent.ALL],
+                'build_profiles': None,
                 'distribution': None,
                 'environment': environment.pk,
                 'host_architecture': 'all',
@@ -618,10 +665,11 @@ class PlaygroundTest(django.test.TestCase, TestCase):
             },
         )
         self.assertEqual(
-            wr.dynamic_task_data,
+            SbuildDynamicData.parse_obj(wr.dynamic_task_data),
             SbuildDynamicData(
                 environment_id=environment.pk,
                 input_source_artifact_id=source.pk,
+                binnmu_maintainer="Debusine <noreply@debusine-playground.net>",
             ),
         )
         self.assertIsNone(wr.parent)
@@ -691,3 +739,37 @@ class PlaygroundTest(django.test.TestCase, TestCase):
                 step="build-all",
             ),
         )
+
+    def test_create_workflow(self) -> None:
+        """Test create_workflow."""
+        playground = Playground()
+        workflow = playground.create_workflow()
+
+        self.assertTrue(workflow.is_workflow)
+
+    def test_create_work_request(self) -> None:
+        """Test create_work_request return a saved work request."""
+        playground = Playground()
+        work_request = playground.create_work_request()
+        work_request.refresh_from_db()
+        self.assertIsInstance(work_request, WorkRequest)
+        self.assertEqual(
+            work_request.workspace, playground.get_default_workspace()
+        )
+
+    def test_create_work_request_use_created_by_user(self) -> None:
+        """Test create_work_request use created_by Token."""
+        playground = Playground()
+        user = playground.create_user("testuser")
+        work_request = playground.create_work_request(created_by=user)
+        work_request.refresh_from_db()
+        self.assertEqual(work_request.created_by, user)
+
+    def test_create_work_request_expired(self) -> None:
+        """Test create_work_request expired=True."""
+        playground = Playground()
+        work_request = playground.create_work_request(expired=True)
+
+        self.assertIsNotNone(work_request.expire_at)
+        assert work_request.expire_at is not None
+        self.assertLess(work_request.expire_at, timezone.now())
