@@ -9,16 +9,21 @@
 
 """Unit tests for DebianPackageBuildLogsManager."""
 
+from django.db.models import Q
+
 from debusine.artifacts.models import (
     ArtifactCategory,
     BareDataCategory,
     CollectionCategory,
 )
-from debusine.db.models import Collection, CollectionItem, default_workspace
+from debusine.client.models import LookupChildType
+from debusine.db.models import CollectionItem, default_workspace
 from debusine.server.collections import (
     DebianPackageBuildLogsManager,
     ItemAdditionError,
 )
+from debusine.server.collections.lookup import LookupResult, lookup_multiple
+from debusine.tasks.models import LookupMultiple
 from debusine.test.django import TestCase
 
 
@@ -27,12 +32,11 @@ class DebianPackageBuildLogsManagerTests(TestCase):
 
     def setUp(self) -> None:
         """Set up tests."""
-        self.user = self.get_test_user()
+        self.user = self.playground.get_default_user()
+        self.workflow = self.playground.create_work_request(task_name="noop")
         self.workspace = default_workspace()
-        self.collection = Collection.objects.create(
-            name="debian",
-            category=CollectionCategory.PACKAGE_BUILD_LOGS,
-            workspace=self.workspace,
+        self.collection = self.workspace.get_singleton_collection(
+            user=self.user, category=CollectionCategory.PACKAGE_BUILD_LOGS
         )
         self.manager = DebianPackageBuildLogsManager(collection=self.collection)
 
@@ -43,7 +47,9 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "Adding to debian:package-build-logs requires data",
         ):
             self.manager.add_bare_data(
-                BareDataCategory.PACKAGE_BUILD_LOG, user=self.user
+                BareDataCategory.PACKAGE_BUILD_LOG,
+                user=self.user,
+                workflow=self.workflow,
             )
 
     def test_do_add_bare_data_raise_item_addition_error(self) -> None:
@@ -57,14 +63,20 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "srcpkg_version": "1.0-1",
         }
         self.manager.add_bare_data(
-            BareDataCategory.PACKAGE_BUILD_LOG, user=self.user, data=data
+            BareDataCategory.PACKAGE_BUILD_LOG,
+            user=self.user,
+            workflow=self.workflow,
+            data=data,
         )
 
         with self.assertRaisesRegex(
             ItemAdditionError, "db_collectionitem_unique_active_name"
         ):
             self.manager.add_bare_data(
-                BareDataCategory.PACKAGE_BUILD_LOG, user=self.user, data=data
+                BareDataCategory.PACKAGE_BUILD_LOG,
+                user=self.user,
+                workflow=self.workflow,
+                data=data,
             )
 
     def test_do_add_bare_data_different_work_request_ids(self) -> None:
@@ -81,6 +93,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             self.manager.add_bare_data(
                 BareDataCategory.PACKAGE_BUILD_LOG,
                 user=self.user,
+                workflow=self.workflow,
                 data={**data, "work_request_id": work_request_id},
             )
             for work_request_id in (1, 2)
@@ -105,12 +118,16 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "srcpkg_version": "1.0-1",
         }
         item_old = self.manager.add_bare_data(
-            BareDataCategory.PACKAGE_BUILD_LOG, user=self.user, data=data
+            BareDataCategory.PACKAGE_BUILD_LOG,
+            user=self.user,
+            workflow=self.workflow,
+            data=data,
         )
 
         item_new = self.manager.add_bare_data(
             BareDataCategory.PACKAGE_BUILD_LOG,
             user=self.user,
+            workflow=self.workflow,
             data=data,
             replace=True,
         )
@@ -120,6 +137,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
         self.assertEqual(item_old.data, data)
         self.assertEqual(item_old.removed_by_user, self.user)
+        self.assertEqual(item_old.removed_by_workflow, self.workflow)
         self.assertIsNotNone(item_old.removed_at)
         self.assertEqual(item_new.name, "debian_bookworm_amd64_hello_1.0-1_1")
         self.assertEqual(item_new.child_type, CollectionItem.Types.BARE)
@@ -140,6 +158,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         item = self.manager.add_bare_data(
             BareDataCategory.PACKAGE_BUILD_LOG,
             user=self.user,
+            workflow=self.workflow,
             data=data,
             replace=True,
         )
@@ -161,14 +180,18 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         item = self.manager.add_bare_data(
             BareDataCategory.PACKAGE_BUILD_LOG,
             user=self.user,
+            workflow=self.workflow,
             data=data,
             replace=True,
         )
 
-        self.manager.remove_bare_data(item.name, user=self.user)
+        self.manager.remove_bare_data(
+            item.name, user=self.user, workflow=self.workflow
+        )
 
         item.refresh_from_db()
         self.assertEqual(item.removed_by_user, self.user)
+        self.assertEqual(item.removed_by_workflow, self.workflow)
         self.assertIsNotNone(item.removed_at)
 
     def test_do_add_artifact_no_variables(self) -> None:
@@ -181,11 +204,13 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             ItemAdditionError,
             "Adding to debian:package-build-logs requires variables",
         ):
-            self.manager.add_artifact(artifact, user=self.user)
+            self.manager.add_artifact(
+                artifact, user=self.user, workflow=self.workflow
+            )
 
     def test_do_add_artifact_raise_item_addition_error(self) -> None:
         """`do_add_artifact` raises an error for duplicate names."""
-        work_request = self.create_work_request()
+        work_request = self.playground.create_work_request()
         artifact_1, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG,
             data={"source": "hello", "version": "1.0-1"},
@@ -200,13 +225,18 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "codename": "bookworm",
             "architecture": "amd64",
         }
-        self.manager.add_artifact(artifact_1, user=self.user, variables=data)
+        self.manager.add_artifact(
+            artifact_1, user=self.user, workflow=self.workflow, variables=data
+        )
 
         with self.assertRaisesRegex(
             ItemAdditionError, "db_collectionitem_unique_active_name"
         ):
             self.manager.add_artifact(
-                artifact_2, user=self.user, variables=data
+                artifact_2,
+                user=self.user,
+                workflow=self.workflow,
+                variables=data,
             )
 
     def test_do_add_artifact_override_srcpkg_name_version(self) -> None:
@@ -216,7 +246,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         This isn't very useful in practice, but it allows its interface to
         be more like `do_add_bare_data`.
         """
-        work_request = self.create_work_request()
+        work_request = self.playground.create_work_request()
         artifact, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG,
             data={"source": "hello", "version": "1.0-1"},
@@ -231,7 +261,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         }
 
         item = self.manager.add_artifact(
-            artifact, user=self.user, variables=data
+            artifact, user=self.user, workflow=self.workflow, variables=data
         )
 
         self.assertEqual(
@@ -241,12 +271,12 @@ class DebianPackageBuildLogsManagerTests(TestCase):
 
     def test_do_add_artifact_different_work_request_ids(self) -> None:
         """Adding artifacts with different work request IDs is OK."""
-        work_request_1 = self.create_work_request()
+        work_request_1 = self.playground.create_work_request()
         artifact_1, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG,
             data={"source": "hello", "version": "1.0-1"},
         )
-        work_request_2 = self.create_work_request()
+        work_request_2 = self.playground.create_work_request()
         artifact_2, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG,
             data={"source": "hello", "version": "1.0-1"},
@@ -261,6 +291,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             self.manager.add_artifact(
                 artifact_1,
                 user=self.user,
+                workflow=self.workflow,
                 variables={**data, "work_request_id": work_request.id},
             )
             for artifact, work_request in (
@@ -279,7 +310,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
 
     def test_do_add_artifact_replace_bare_data(self) -> None:
         """`do_add_artifact` can replace an existing bare data item."""
-        work_request = self.create_work_request()
+        work_request = self.playground.create_work_request()
         data = {
             "work_request_id": work_request.id,
             "vendor": "debian",
@@ -289,7 +320,10 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "srcpkg_version": "1.0-1",
         }
         item_old = self.manager.add_bare_data(
-            BareDataCategory.PACKAGE_BUILD_LOG, user=self.user, data=data
+            BareDataCategory.PACKAGE_BUILD_LOG,
+            user=self.user,
+            workflow=self.workflow,
+            data=data,
         )
         artifact_new, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG,
@@ -297,7 +331,11 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         )
 
         item_new = self.manager.add_artifact(
-            artifact_new, user=self.user, variables=data, replace=True
+            artifact_new,
+            user=self.user,
+            workflow=self.workflow,
+            variables=data,
+            replace=True,
         )
 
         item_old.refresh_from_db()
@@ -308,6 +346,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
         self.assertEqual(item_old.data, data)
         self.assertEqual(item_old.removed_by_user, self.user)
+        self.assertEqual(item_old.removed_by_workflow, self.workflow)
         self.assertIsNotNone(item_old.removed_at)
         self.assertEqual(
             item_new.name,
@@ -321,7 +360,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         """`do_add_artifact` can replace an existing artifact."""
         worker_1 = self.playground.create_worker()
         worker_2 = self.playground.create_worker()
-        work_request = self.create_work_request(worker=worker_1)
+        work_request = self.playground.create_work_request(worker=worker_1)
         artifact_old, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
         )
@@ -337,13 +376,14 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "srcpkg_version": "1.0-1",
         }
         item_old = self.manager.add_artifact(
-            artifact_old, user=self.user, variables=data
+            artifact_old, user=self.user, workflow=self.workflow, variables=data
         )
 
         work_request.assign_worker(worker_2)
         item_new = self.manager.add_artifact(
             artifact_new,
             user=self.user,
+            workflow=self.workflow,
             variables=data,
             replace=True,
         )
@@ -356,6 +396,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         self.assertEqual(item_old.artifact, artifact_old)
         self.assertEqual(item_old.data, {**data, "worker": worker_1.name})
         self.assertEqual(item_old.removed_by_user, self.user)
+        self.assertEqual(item_old.removed_by_workflow, self.workflow)
         self.assertIsNotNone(item_old.removed_at)
         self.assertEqual(
             item_new.name,
@@ -367,7 +408,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
 
     def test_do_add_artifact_replace_nonexistent(self) -> None:
         """Replacing a nonexistent artifact is allowed."""
-        work_request = self.create_work_request()
+        work_request = self.playground.create_work_request()
         artifact, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
         )
@@ -383,6 +424,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
         item = self.manager.add_artifact(
             artifact,
             user=self.user,
+            workflow=self.workflow,
             variables=data,
             replace=True,
         )
@@ -395,7 +437,7 @@ class DebianPackageBuildLogsManagerTests(TestCase):
 
     def test_do_remove_artifact(self) -> None:
         """`do_remove_artifact` removes the item."""
-        work_request = self.create_work_request()
+        work_request = self.playground.create_work_request()
         artifact, _ = self.create_artifact(
             category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
         )
@@ -408,11 +450,175 @@ class DebianPackageBuildLogsManagerTests(TestCase):
             "srcpkg_version": "1.0-1",
         }
         item = self.manager.add_artifact(
-            artifact, user=self.user, variables=data, replace=True
+            artifact,
+            user=self.user,
+            workflow=self.workflow,
+            variables=data,
+            replace=True,
         )
 
-        self.manager.remove_artifact(artifact, user=self.user)
+        self.manager.remove_artifact(
+            artifact, user=self.user, workflow=self.workflow
+        )
 
         item.refresh_from_db()
         self.assertEqual(item.removed_by_user, self.user)
+        self.assertEqual(item.removed_by_workflow, self.workflow)
         self.assertIsNotNone(item.removed_at)
+
+    def test_do_lookup_filter_unexpected_format(self) -> None:
+        """`do_lookup_filter` raises LookupError for an unexpected format."""
+        with self.assertRaisesRegex(
+            LookupError, r'^Unexpected lookup filter format: "invalid"'
+        ):
+            self.manager.lookup_filter(
+                "invalid", "foo", workspace=self.workspace, user=self.user
+            )
+
+    def test_do_lookup_filter_same_work_request_single(self) -> None:
+        """`do_lookup_filter`: `same_work_request` with a single lookup."""
+        work_requests = [
+            self.playground.create_work_request() for _ in range(2)
+        ]
+        upload, _ = self.playground.create_artifact(
+            category=ArtifactCategory.UPLOAD, work_request=work_requests[0]
+        )
+        build_logs = [
+            self.playground.create_artifact(
+                category=ArtifactCategory.PACKAGE_BUILD_LOG,
+                work_request=work_request,
+            )[0]
+            for work_request in work_requests
+        ]
+        items = [
+            self.manager.add_artifact(
+                build_log,
+                user=self.user,
+                variables={
+                    "work_request_id": build_log.created_by_work_request_id,
+                    "vendor": "debian",
+                    "codename": "bookworm",
+                    "architecture": "amd64",
+                    "srcpkg_name": "hello",
+                    "srcpkg_version": "1.0-1",
+                },
+            )
+            for build_log in build_logs
+        ]
+        subordinate_lookup = f"{upload.id}@artifacts"
+
+        condition = self.manager.lookup_filter(
+            "same_work_request",
+            subordinate_lookup,
+            workspace=self.workspace,
+            user=self.user,
+        )
+        self.assertEqual(
+            condition, Q(data__work_request_id__in={work_requests[0].id})
+        )
+        self.assertQuerySetEqual(
+            CollectionItem.active_objects.filter(
+                parent_collection=self.collection
+            ).filter(condition),
+            [items[0]],
+        )
+        self.assertCountEqual(
+            lookup_multiple(
+                LookupMultiple.parse_obj(
+                    {
+                        "collection": self.collection.id,
+                        "lookup__same_work_request": subordinate_lookup,
+                    }
+                ),
+                self.workspace,
+                user=self.user,
+                expect_type=LookupChildType.ARTIFACT,
+            ),
+            [
+                LookupResult(
+                    result_type=CollectionItem.Types.ARTIFACT,
+                    collection_item=items[0],
+                    artifact=build_logs[0],
+                )
+            ],
+        )
+
+    def test_do_lookup_filter_same_work_request_multiple(self) -> None:
+        """`do_lookup_filter`: `same_work_request` with a multiple lookup."""
+        work_requests = [
+            self.playground.create_work_request() for _ in range(3)
+        ]
+        uploads = [
+            self.playground.create_artifact(
+                category=ArtifactCategory.UPLOAD, work_request=work_request
+            )[0]
+            for work_request in work_requests[:2]
+        ]
+        build_logs = [
+            self.playground.create_artifact(
+                category=ArtifactCategory.PACKAGE_BUILD_LOG,
+                work_request=work_request,
+            )[0]
+            for work_request in work_requests
+        ]
+        items = [
+            self.manager.add_artifact(
+                build_log,
+                user=self.user,
+                variables={
+                    "work_request_id": build_log.created_by_work_request_id,
+                    "vendor": "debian",
+                    "codename": "bookworm",
+                    "architecture": "amd64",
+                    "srcpkg_name": "hello",
+                    "srcpkg_version": "1.0-1",
+                },
+            )
+            for build_log in build_logs
+        ]
+        subordinate_lookup = LookupMultiple.parse_obj(
+            [f"{upload.id}@artifacts" for upload in uploads]
+        )
+
+        condition = self.manager.lookup_filter(
+            "same_work_request",
+            subordinate_lookup,
+            workspace=self.workspace,
+            user=self.user,
+        )
+        self.assertEqual(
+            condition,
+            Q(
+                data__work_request_id__in={
+                    work_request.id for work_request in work_requests[:2]
+                }
+            ),
+        )
+        self.assertQuerySetEqual(
+            CollectionItem.active_objects.filter(
+                parent_collection=self.collection
+            ).filter(condition),
+            items[:2],
+            ordered=False,
+        )
+        self.assertCountEqual(
+            lookup_multiple(
+                LookupMultiple.parse_obj(
+                    {
+                        "collection": self.collection.id,
+                        "lookup__same_work_request": subordinate_lookup,
+                    }
+                ),
+                self.workspace,
+                user=self.user,
+                expect_type=LookupChildType.ARTIFACT,
+            ),
+            [
+                LookupResult(
+                    result_type=CollectionItem.Types.ARTIFACT,
+                    collection_item=item,
+                    artifact=build_log,
+                )
+                for item, build_log in zip(items[:2], build_logs[:2])
+            ],
+        )
