@@ -17,12 +17,7 @@ from unittest import mock
 from django.test import TestCase
 from django.utils import timezone
 
-from debusine.db.models import (
-    Artifact,
-    FileInArtifact,
-    FileStore,
-    FileUpload,
-)
+from debusine.db.models import Artifact, FileInArtifact, FileStore, FileUpload
 from debusine.django.management.tests import call_command
 from debusine.project import settings
 from debusine.server.management.commands.monthly_cleanup import Command
@@ -261,7 +256,7 @@ class MonthlyCleanupCommandTests(TestHelpersMixin, TestCase):
         self.assertEqual(exitcode, 0)
 
         # Check that the FileUpload row was not deleted
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             FileUpload.objects.filter(id=file_upload.id), [file_upload]
         )
 
@@ -284,7 +279,7 @@ class MonthlyCleanupCommandTests(TestHelpersMixin, TestCase):
         self.assertEqual(exitcode, 0)
 
         # Check that the FileUpload row was not deleted
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             FileUpload.objects.filter(id=file_upload.id), []
         )
 
@@ -306,7 +301,7 @@ class MonthlyCleanupCommandTests(TestHelpersMixin, TestCase):
         call_command("monthly_cleanup")
 
         # FileUpload did not get deleted: recent activity
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             FileUpload.objects.filter(id=file_upload.id), [file_upload]
         )
 
