@@ -8,6 +8,8 @@
 # contained in the LICENSE file.
 
 """Tests for the work request views."""
+
+import re
 from base64 import b64encode
 from contextlib import contextmanager
 from datetime import datetime
@@ -22,23 +24,21 @@ from django.test import override_settings
 from django.urls import reverse
 from django.utils import timezone
 from django.utils.http import urlencode
-
 from rest_framework import status
 
+from debusine.artifacts.models import ArtifactCategory
 from debusine.db.models import (
     Artifact,
+    ArtifactRelation,
     DEFAULT_WORKSPACE_NAME,
     Token,
     WorkRequest,
     Worker,
 )
-from debusine.server.serializers import (
-    WorkRequestSerializer,
-)
-from debusine.server.views.work_requests import (
-    WorkRequestPagination,
-)
+from debusine.server.serializers import WorkRequestSerializer
+from debusine.server.views.work_requests import WorkRequestPagination
 from debusine.server.workflows.models import WorkRequestManualUnblockAction
+from debusine.tasks.models import TaskTypes
 from debusine.test.django import TestCase
 from debusine.test.utils import date_time_to_isoformat_rest_framework
 
@@ -50,7 +50,10 @@ class WorkRequestTestCase(TestCase):
     """Helper methods for tests with WorkRequest as response."""
 
     def check_response_for_work_request(
-        self, response: "_MonkeyPatchedWSGIResponse", work_request: WorkRequest
+        self,
+        response: "_MonkeyPatchedWSGIResponse",
+        work_request: WorkRequest,
+        dynamic_task_data: dict[str, Any] | None = None,
     ):
         """Ensure the json data corresponds to the supplied work request."""
 
@@ -77,7 +80,11 @@ class WorkRequestTestCase(TestCase):
                 'worker': work_request.worker_id,
                 'task_type': work_request.task_type,
                 'task_data': work_request.task_data,
-                'dynamic_task_data': None,
+                'dynamic_task_data': (
+                    dynamic_task_data
+                    if work_request.dynamic_task_data is None
+                    else work_request.dynamic_task_data
+                ),
                 'priority_base': work_request.priority_base,
                 'priority_adjustment': work_request.priority_adjustment,
                 'workflow_data': work_request.workflow_data_json,
@@ -194,11 +201,11 @@ class WorkRequestViewTests(WorkRequestTestCase):
         """Set up common data."""
         super().setUp()
         self.worker_01 = Worker.objects.create_with_fqdn(
-            "worker-01", self.create_token_enabled()
+            "worker-01", self.playground.create_token_enabled()
         )
 
         # This token would be used by a debusine client using the API
-        self.token = self.create_token_enabled(with_user=True)
+        self.token = self.playground.create_token_enabled(with_user=True)
 
     @classmethod
     def setUpTestData(cls) -> None:
@@ -208,7 +215,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
 
     def create_test_work_request(self):
         """Create a work request for testing."""
-        self.work_request_01 = self.create_work_request(
+        self.work_request_01 = self.playground.create_work_request(
             worker=self.worker_01,
             task_name='sbuild',
             task_data={'architecture': 'amd64'},
@@ -226,7 +233,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
             email="superuser@mail.none",
             password="testsuperpass",
         )
-        return self.create_token_enabled(user=user)
+        return self.playground.create_token_enabled(user=user)
 
     def post_work_requests(
         self,
@@ -240,7 +247,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         response = self.client.post(
             reverse('api:work-requests'),
             data=work_request_serialized,
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -248,7 +255,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
 
     def test_check_permissions_get_token_must_be_enabled(self):
         """An enabled token is needed to request a WorkRequest."""
-        token = self.create_token_enabled()
+        token = self.playground.create_token_enabled()
         token.disable()
         token.save()
         self.create_test_work_request()
@@ -258,14 +265,14 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 "api:work-request-detail",
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
         )
 
         self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
 
     def test_check_permissions_post_token_must_have_user_associated(self):
         """The token must have a user associated for creating a WorkRequest."""
-        token = self.create_token_enabled()
+        token = self.playground.create_token_enabled()
 
         work_request_serialized = {
             "task_name": "sbuild",
@@ -275,7 +282,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         response = self.client.post(
             reverse('api:work-requests'),
             data=work_request_serialized,
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -290,7 +297,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 'api:work-request-detail',
                 kwargs={'work_request_id': self.work_request_01.id},
             ),
-            HTTP_TOKEN=self.worker_01.token.key,
+            headers={"token": self.worker_01.token.key},
         )
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
@@ -306,7 +313,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 'api:work-request-detail',
                 kwargs={'work_request_id': max_id + 1},
             ),
-            HTTP_TOKEN=self.worker_01.token.key,
+            headers={"token": self.worker_01.token.key},
         )
 
         self.assertEqual(
@@ -543,7 +550,8 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_list_work_request_empty(self):
         """Return an empty list if there are no work requests."""
         response = self.client.get(
-            reverse('api:work-requests'), HTTP_TOKEN=self.worker_01.token.key
+            reverse('api:work-requests'),
+            headers={"token": self.worker_01.token.key},
         )
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
@@ -553,7 +561,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         """Return a single page if there are only a few work requests."""
         work_requests = []
         for _ in range(3):
-            work_request = self.create_work_request(
+            work_request = self.playground.create_work_request(
                 worker=self.worker_01,
                 task_name='sbuild',
                 task_data={'architecture': 'amd64'},
@@ -565,7 +573,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         with mock.patch.object(WorkRequestPagination, 'page_size', 5):
             response = self.client.get(
                 reverse('api:work-requests'),
-                HTTP_TOKEN=self.worker_01.token.key,
+                headers={"token": self.worker_01.token.key},
             )
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
@@ -575,7 +583,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         """Return multiple pages if there are many work requests."""
         work_requests = []
         for _ in range(10):
-            work_request = self.create_work_request(
+            work_request = self.playground.create_work_request(
                 worker=self.worker_01,
                 task_name='sbuild',
                 task_data={'architecture': 'amd64'},
@@ -587,7 +595,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         with mock.patch.object(WorkRequestPagination, 'page_size', 5):
             response = self.client.get(
                 reverse('api:work-requests'),
-                HTTP_TOKEN=self.worker_01.token.key,
+                headers={"token": self.worker_01.token.key},
             )
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
@@ -598,7 +606,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
         with mock.patch.object(WorkRequestPagination, 'page_size', 5):
             response = self.client.get(
                 response.json()['next'],
-                HTTP_TOKEN=self.worker_01.token.key,
+                headers={"token": self.worker_01.token.key},
             )
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
@@ -609,8 +617,8 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_patch_work_request_success(self):
         """Patch a work request and return its new representation."""
         self.create_test_work_request()
-        token = self.create_token_enabled(with_user=True)
-        self.add_user_permission(
+        token = self.playground.create_token_enabled(with_user=True)
+        self.playground.add_user_permission(
             token.user, WorkRequest, "manage_workrequest_priorities"
         )
 
@@ -620,7 +628,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
             data={"priority_adjustment": 100},
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -632,7 +640,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_patch_work_request_empty(self):
         """An empty patch to a work request does nothing."""
         self.create_test_work_request()
-        token = self.create_token_enabled(with_user=True)
+        token = self.playground.create_token_enabled(with_user=True)
 
         response = self.client.patch(
             reverse(
@@ -640,7 +648,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
             data={},
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -672,7 +680,9 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_patch_work_request_own_request_negative_adjustment(self):
         """Users can set negative adjustments on their own work requests."""
         self.create_test_work_request()
-        token = self.create_token_enabled(user=self.work_request_01.created_by)
+        token = self.playground.create_token_enabled(
+            user=self.work_request_01.created_by
+        )
 
         response = self.client.patch(
             reverse(
@@ -680,7 +690,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
             data={"priority_adjustment": -100},
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -692,7 +702,9 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_patch_work_request_own_request_positive_adjustment(self):
         """Users cannot set positive adjustments on their own work requests."""
         self.create_test_work_request()
-        token = self.create_token_enabled(user=self.work_request_01.created_by)
+        token = self.playground.create_token_enabled(
+            user=self.work_request_01.created_by
+        )
 
         response = self.client.patch(
             reverse(
@@ -700,7 +712,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
             data={"priority_adjustment": 100},
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -713,7 +725,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_patch_work_request_without_permission(self):
         """Setting positive adjustments requires a special permission."""
         self.create_test_work_request()
-        token = self.create_token_enabled(with_user=True)
+        token = self.playground.create_token_enabled(with_user=True)
 
         response = self.client.patch(
             reverse(
@@ -721,7 +733,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
             data={"priority_adjustment": 100},
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -734,8 +746,8 @@ class WorkRequestViewTests(WorkRequestTestCase):
     def test_patch_work_request_disallowed_property(self):
         """Only certain properties of work requests may be patched."""
         self.create_test_work_request()
-        token = self.create_token_enabled(with_user=True)
-        self.add_user_permission(
+        token = self.playground.create_token_enabled(with_user=True)
+        self.playground.add_user_permission(
             token.user, WorkRequest, "manage_workrequest_priorities"
         )
 
@@ -745,7 +757,7 @@ class WorkRequestViewTests(WorkRequestTestCase):
                 kwargs={"work_request_id": self.work_request_01.id},
             ),
             data={"priority_base": 100},
-            HTTP_TOKEN=token.key,
+            headers={"token": token.key},
             content_type="application/json",
         )
 
@@ -773,9 +785,9 @@ class WorkRequestRetryViewTests(TestCase):
     def setUpTestData(cls):
         """Set up common data."""
         super().setUpTestData()
-        cls.work_request = cls.create_work_request(task_name="noop")
+        cls.work_request = cls.playground.create_work_request(task_name="noop")
         # This token would be used by a debusine client using the API
-        cls.token = cls.create_token_enabled(with_user=True)
+        cls.token = cls.playground.create_token_enabled(with_user=True)
 
     def post(
         self,
@@ -795,7 +807,7 @@ class WorkRequestRetryViewTests(TestCase):
                 },
             ),
             data={},
-            HTTP_TOKEN=(token or self.token).key,
+            headers={"token": (token or self.token).key},
             content_type="application/json",
         )
 
@@ -864,11 +876,11 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def setUpTestData(cls) -> None:
         """Set up common data."""
         super().setUpTestData()
-        template = cls.create_workflow_template("test", "noop")
+        template = cls.playground.create_workflow_template("test", "noop")
         cls.workflow = WorkRequest.objects.create_workflow(
             template=template,
             data={},
-            created_by=cls.get_test_user(),
+            created_by=cls.playground.get_default_user(),
             status=WorkRequest.Statuses.RUNNING,
         )
         cls.work_request = WorkRequest.objects.create_synchronization_point(
@@ -879,7 +891,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
         cls.work_request.unblock_strategy = WorkRequest.UnblockStrategy.MANUAL
         cls.work_request.save()
         # This token would be used by a debusine client using the API
-        cls.token = cls.create_token_enabled(with_user=True)
+        cls.token = cls.playground.create_token_enabled(with_user=True)
 
     def post(
         self,
@@ -906,7 +918,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
                 },
             ),
             data=data,
-            HTTP_TOKEN=(token or self.token).key,
+            headers={"token": (token or self.token).key},
             content_type="application/json",
         )
 
@@ -942,7 +954,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_neither_notes_nor_action(self) -> None:
         """At least one of notes and action must be set."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
         response = self.post(work_request_id=self.work_request.id)
@@ -956,7 +968,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_work_request_not_found(self) -> None:
         """Return HTTP 404 if the WorkRequest id does not exist."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
         max_id = WorkRequest.objects.aggregate(Max("id"))["id__max"]
@@ -973,7 +985,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_not_manual(self) -> None:
         """The work request must have the manual unblock strategy."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
         self.work_request.unblock_strategy = WorkRequest.UnblockStrategy.DEPS
@@ -989,10 +1001,10 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_not_part_of_workflow(self) -> None:
         """The work request must be part of a workflow."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
-        work_request = self.create_work_request(
+        work_request = self.playground.create_work_request(
             task_name="noop",
             status=WorkRequest.Statuses.BLOCKED,
             unblock_strategy=WorkRequest.UnblockStrategy.MANUAL,
@@ -1008,7 +1020,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_not_blocked(self) -> None:
         """The work request must currently be blocked."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
         self.work_request.status = WorkRequest.Statuses.PENDING
@@ -1023,7 +1035,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_accept(self) -> None:
         """Accept a blocked work request."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
 
@@ -1051,7 +1063,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_reject(self) -> None:
         """Reject a blocked work request."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
 
@@ -1080,7 +1092,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_record_notes_only(self) -> None:
         """Record notes on a blocked work request."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
 
@@ -1105,7 +1117,7 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
     def test_post_reject_retry_accept(self) -> None:
         """Reject a blocked work request, then retry it, then accept it."""
         assert self.token.user is not None
-        self.add_user_permission(
+        self.playground.add_user_permission(
             self.token.user, WorkRequest, "change_workrequest"
         )
 
@@ -1151,3 +1163,422 @@ class WorkRequestUnblockViewTests(WorkRequestTestCase):
         )
         assert isinstance(response, HttpResponse)
         self.check_response_for_work_request(response, retried_work_request)
+
+
+class WorkRequestExternalDebsignViewTests(WorkRequestTestCase):
+    """Tests for :py:class:`WorkRequestExternalDebsignView`."""
+
+    unsigned_files: ClassVar[dict[str, bytes]]
+    unsigned: ClassVar[Artifact]
+    work_request: ClassVar[WorkRequest]
+    token: ClassVar[Token]
+
+    @classmethod
+    def setUpTestData(cls) -> None:
+        """Set up common data."""
+        super().setUpTestData()
+        cls.unsigned_files = {
+            "foo_1.0.orig.tar.gz": b"orig.tar.gz",
+            "foo_1.0.dsc": b"unsigned dsc",
+            "foo_1.0_source.changes": b"unsigned changes",
+        }
+        cls.unsigned, _ = cls.playground.create_artifact(
+            paths=cls.unsigned_files,
+            category=ArtifactCategory.UPLOAD,
+            create_files=True,
+            skip_add_files_in_store=True,
+        )
+        cls.work_request = cls.playground.create_work_request(
+            mark_running=True,
+            task_type=TaskTypes.WAIT,
+            task_name="externaldebsign",
+            task_data={"unsigned": f"{cls.unsigned.id}@artifacts"},
+        )
+        cls.token = cls.playground.create_token_enabled(with_user=True)
+
+    def get(
+        self, *, work_request_id: int | None = None, token: Token | None = None
+    ) -> "_MonkeyPatchedWSGIResponse":
+        """
+        Get the work request from api:work-request-external-debsign.
+
+        :param work_request_id: ID of the work request to get, defaulting to
+          the ID of `self.work_request`.
+        :param token: authentication token to send, defaulting to
+          `self.token`.
+        """
+        return self.client.get(
+            reverse(
+                "api:work-request-external-debsign",
+                kwargs={
+                    "work_request_id": (
+                        self.work_request.id
+                        if work_request_id is None
+                        else work_request_id
+                    )
+                },
+            ),
+            headers={"token": (token or self.token).key},
+        )
+
+    def post(
+        self,
+        *,
+        work_request_id: int | None = None,
+        token: Token | None = None,
+        signed: Artifact,
+    ) -> "_MonkeyPatchedWSGIResponse":
+        """
+        Post a signed artifact to api:work-request-external-debsign.
+
+        :param work_request_id: ID of the work request to post to,
+          defaulting to the ID of `self.work_request`.
+        :param token: authentication token to send, defaulting to
+          `self.token`.
+        :param signed: the signed artifact to post.
+        """
+        return self.client.post(
+            reverse(
+                "api:work-request-external-debsign",
+                kwargs={
+                    "work_request_id": (
+                        self.work_request.id
+                        if work_request_id is None
+                        else work_request_id
+                    )
+                },
+            ),
+            data={"signed_artifact": signed.id},
+            headers={"token": (token or self.token).key},
+            content_type="application/json",
+        )
+
+    @classmethod
+    def create_signed_artifact(
+        cls,
+        paths: list[str] | dict[str, bytes] | None = None,
+        category: ArtifactCategory = ArtifactCategory.UPLOAD,
+    ) -> Artifact:
+        """
+        Create a signed upload artifact.
+
+        :param paths: list of paths to create (with random data), or
+          dictionary mapping paths to create to their data; defaults to a
+          structurally-valid signed upload
+        :param category: the upload artifact category, defaulting to
+          `debian:upload`
+        """
+        signed, _ = cls.playground.create_artifact(
+            paths=(
+                paths
+                or {
+                    **cls.unsigned_files,
+                    "foo_1.0.dsc": b"signed dsc",
+                    "foo_1.0_source.changes": b"signed changes",
+                }
+            ),
+            category=category,
+            create_files=True,
+            skip_add_files_in_store=True,
+        )
+        return signed
+
+    def test_get_token_must_be_enabled(self) -> None:
+        """A `GET` request needs an enabled token."""
+        self.token.disable()
+        self.token.save()
+
+        response = self.get()
+
+        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
+
+    def test_get_token_must_have_user_associated(self) -> None:
+        """A `GET` request needs a token with a user."""
+        self.token.user = None
+        self.token.save()
+
+        response = self.get()
+
+        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
+
+    def test_get_work_request_not_found(self) -> None:
+        """A `GET` request returns 404 if the work request does not exist."""
+        max_id = WorkRequest.objects.aggregate(Max("id"))["id__max"]
+
+        response = self.get(work_request_id=max_id + 1)
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="No WorkRequest matches the given query.",
+            status_code=status.HTTP_404_NOT_FOUND,
+        )
+
+    def test_get_wrong_work_request(self) -> None:
+        """A `GET` request requires the right kind of work request."""
+        work_request = self.playground.create_work_request(
+            task_type=TaskTypes.WORKER, task_name="noop"
+        )
+
+        response = self.get(work_request_id=work_request.id)
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern=(
+                "Expected work request to be Wait/ExternalDebsign; got "
+                "Worker/noop"
+            ),
+        )
+
+    def test_get_work_request_not_running(self) -> None:
+        """A `GET` request requires that the work request be running."""
+        self.work_request.mark_aborted()
+
+        response = self.get()
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="Work request is not running. Status: aborted",
+        )
+
+    def test_get_bad_task_data(self) -> None:
+        """A `GET` request fails on bad task data."""
+        self.work_request.task_data = {"unsigned": ""}
+        self.work_request.save()
+
+        response = self.get()
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="Cannot process task data: Empty lookup",
+        )
+
+    def test_get_success(self) -> None:
+        """A successful `GET` request returns the serialized work request."""
+        response = self.get()
+
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+        self.work_request.refresh_from_db()
+        self.check_response_for_work_request(
+            response,
+            self.work_request,
+            dynamic_task_data={"unsigned_id": self.unsigned.id},
+        )
+
+    def test_get_different_user(self) -> None:
+        """A `GET` request does not have to be sent by the WR creator."""
+        user = get_user_model().objects.create_user(
+            username="another", email="another@example.org"
+        )
+        token = self.playground.create_token_enabled(user=user)
+
+        response = self.get(token=token)
+
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+        self.work_request.refresh_from_db()
+        self.check_response_for_work_request(
+            response,
+            self.work_request,
+            dynamic_task_data={"unsigned_id": self.unsigned.id},
+        )
+
+    def test_post_token_must_be_enabled(self) -> None:
+        """A `POST` request needs an enabled token."""
+        self.token.disable()
+        self.token.save()
+
+        response = self.post(signed=self.create_signed_artifact())
+
+        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
+
+    def test_post_token_must_have_user_associated(self) -> None:
+        """A `POST` request needs a token with a user."""
+        self.token.user = None
+        self.token.save()
+
+        response = self.post(signed=self.create_signed_artifact())
+
+        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
+
+    def test_post_work_request_not_found(self) -> None:
+        """A `POST` request returns 404 if the work request does not exist."""
+        max_id = WorkRequest.objects.aggregate(Max("id"))["id__max"]
+
+        response = self.post(
+            work_request_id=max_id + 1, signed=self.create_signed_artifact()
+        )
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="No WorkRequest matches the given query.",
+            status_code=status.HTTP_404_NOT_FOUND,
+        )
+
+    def test_post_wrong_work_request(self) -> None:
+        """A `POST` request requires the right kind of work request."""
+        work_request = self.playground.create_work_request(
+            task_type=TaskTypes.WORKER, task_name="noop"
+        )
+
+        response = self.post(
+            work_request_id=work_request.id,
+            signed=self.create_signed_artifact(),
+        )
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern=(
+                "Expected work request to be Wait/ExternalDebsign; got "
+                "Worker/noop"
+            ),
+        )
+
+    def test_post_work_request_not_running(self) -> None:
+        """A `POST` request requires that the work request be running."""
+        self.work_request.mark_aborted()
+
+        response = self.post(signed=self.create_signed_artifact())
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="Work request is not running. Status: aborted",
+        )
+
+    def test_post_bad_task_data(self) -> None:
+        """A `POST` request fails on bad task data."""
+        self.work_request.task_data = {"unsigned": ""}
+        self.work_request.save()
+
+        response = self.post(signed=self.create_signed_artifact())
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="Cannot process task data: Empty lookup",
+        )
+
+    def test_post_wrong_user(self) -> None:
+        """A `POST` request may only be sent by the user who created the WR."""
+        user = get_user_model().objects.create_user(
+            username="another", email="another@example.org"
+        )
+        token = self.playground.create_token_enabled(user=user)
+
+        response = self.post(token=token, signed=self.create_signed_artifact())
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern=(
+                "Only the user who created the work request can provide a "
+                "signature"
+            ),
+            status_code=status.HTTP_403_FORBIDDEN,
+        )
+
+    def test_post_wrong_category(self) -> None:
+        """A `POST` request checks the category of the signed artifact."""
+        signed = self.create_signed_artifact(
+            category=ArtifactCategory.SOURCE_PACKAGE
+        )
+
+        response = self.post(signed=signed)
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern=(
+                "Expected signed artifact of category debian:upload; got "
+                "debian:source-package"
+            ),
+        )
+
+    def test_post_signed_artifact_new_files(self) -> None:
+        """The signed artifact may not add new files."""
+        signed = self.create_signed_artifact(
+            paths={
+                **self.unsigned_files,
+                "foo_1.0-1.debian.tar.xz": b"debian.tar.xz",
+            }
+        )
+
+        response = self.post(signed=signed)
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern=re.escape(
+                "Signed upload adds extra files when compared to unsigned "
+                "upload. New paths: ['foo_1.0-1.debian.tar.xz']"
+            ),
+        )
+
+    def test_post_signed_artifact_unchanged_changes_file(self) -> None:
+        """The signed artifact must change the `.changes` file."""
+        signed = self.create_signed_artifact(paths={**self.unsigned_files})
+
+        response = self.post(signed=signed)
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern="Signed upload does not change the .changes file",
+        )
+
+    def test_post_signed_artifact_changed_wrong_files(self) -> None:
+        """The signed artifact may not change unsignable files."""
+        signed = self.create_signed_artifact(
+            paths={
+                **self.unsigned_files,
+                "foo_1.0.orig.tar.gz": b"new orig.tar.gz",
+                "foo_1.0_source.changes": b"signed changes",
+            }
+        )
+
+        response = self.post(signed=signed)
+
+        self.assertResponseProblem(
+            response,
+            "Error",
+            detail_pattern=re.escape(
+                "Signed upload changes more than .changes/.dsc/.buildinfo. "
+                "Changed paths: "
+                "['foo_1.0.orig.tar.gz', 'foo_1.0_source.changes']"
+            ),
+        )
+
+    def test_post_signed_artifact_success(self) -> None:
+        """A successful `POST` request completes and returns 200."""
+        signed = self.create_signed_artifact()
+
+        response = self.post(signed=signed)
+
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+        self.assertQuerySetEqual(
+            ArtifactRelation.objects.filter(artifact=signed).values(
+                "artifact", "target", "type"
+            ),
+            [
+                {
+                    "artifact": signed.id,
+                    "target": self.unsigned.id,
+                    "type": ArtifactRelation.Relations.RELATES_TO,
+                }
+            ],
+        )
+        self.work_request.refresh_from_db()
+        self.check_response_for_work_request(
+            response,
+            self.work_request,
+            dynamic_task_data={"unsigned_id": self.unsigned.id},
+        )
+        self.assertEqual(
+            self.work_request.status, WorkRequest.Statuses.COMPLETED
+        )
+        self.assertEqual(self.work_request.result, WorkRequest.Results.SUCCESS)
