@@ -11,7 +11,7 @@
 
 from unittest import TestCase
 
-from debusine.signing.tasks.noop import SigningNoop
+from debusine.signing.tasks import SigningNoop
 from debusine.tasks import TaskConfigError
 
 
