@@ -17,11 +17,9 @@ from tempfile import TemporaryDirectory
 from typing import Any
 from unittest import mock, skipUnless
 
+import gpg
 from django.test import TestCase, TransactionTestCase, override_settings
 from django.utils import timezone
-
-import gpg
-
 from nacl.public import PrivateKey
 
 try:
@@ -37,6 +35,7 @@ from debusine.artifacts import (
     WorkRequestDebugLogs,
 )
 from debusine.artifacts.models import (
+    ArtifactCategory,
     DebusineSigningOutput,
     EmptyArtifactData,
     SigningResult,
@@ -50,14 +49,14 @@ from debusine.client.models import (
 )
 from debusine.signing.db.models import Key
 from debusine.signing.gnupg import gpg_ephemeral_context
+from debusine.signing.models import SigningMode
+from debusine.signing.tasks import BaseSigningTask, Sign
 from debusine.signing.tasks.models import KeyPurpose, SignDynamicData
-from debusine.signing.tasks.sign import Sign
-from debusine.tasks.models import WorkerType
+from debusine.tasks.models import LookupMultiple, WorkerType
 from debusine.tasks.tests.helper_mixin import FakeTaskDatabase
 from debusine.test import TestHelpersMixin
 from debusine.utils import calculate_hash, is_command_available
 
-
 # Bare minimum that sbsigntool will recognize as a valid PE/COFF image.
 # This is unlikely to work in any other context.
 minimal_pecoff_image = (
@@ -111,7 +110,7 @@ example_url = pydantic.parse_obj_as(pydantic.AnyUrl, "http://example.org/")
 class SignTestMixin:
     """Helpers for signing tests."""
 
-    def mock_debusine(self, task: Sign) -> mock.MagicMock:
+    def mock_debusine(self, task: BaseSigningTask[Any, Any]) -> mock.MagicMock:
         """Create a Debusine mock and configure a task for it."""
         debusine_mock = mock.create_autospec(spec=Debusine)
         task.configure_server_access(debusine_mock)
@@ -154,7 +153,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         """Test the analyze_worker() method."""
         self.mock_is_command_available({"sbsign": True})
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         metadata = task.analyze_worker()
         self.assertEqual(metadata["signing:sign:available:uefi"], True)
@@ -163,7 +162,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         """analyze_worker() handles sbsign not being available."""
         self.mock_is_command_available({"sbsign": False})
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         metadata = task.analyze_worker()
         self.assertEqual(metadata["signing:sign:available:uefi"], False)
@@ -171,7 +170,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
     def test_can_run_on_uefi(self) -> None:
         """can_run_on returns True for UEFI if sbsign is available."""
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         self.assertTrue(
             task.can_run_on(
@@ -186,7 +185,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
     def test_can_run_on_openpgp(self) -> None:
         """can_run_on always returns True for OpenPGP."""
         task = Sign(
-            task_data={"purpose": KeyPurpose.OPENPGP, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.OPENPGP, "unsigned": [2], "key": 1}
         )
         self.assertTrue(
             task.can_run_on(
@@ -200,7 +199,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
     def test_can_run_on_mismatched_task_version(self) -> None:
         """can_run_on returns False for mismatched task versions."""
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         self.assertFalse(
             task.can_run_on(
@@ -215,7 +214,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
     def test_can_run_on_uefi_missing_tool(self) -> None:
         """can_run_on returns False for UEFI if sbsign is not available."""
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         self.assertFalse(
             task.can_run_on(
@@ -232,23 +231,25 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         key_lookup = "bookworm@debian:suite-signing-keys/key:uefi"
         task_db = FakeTaskDatabase(
             single_lookups={
-                # unsigned
-                (2, None): 2,
                 # key
                 (key_lookup, None): 1,
-            }
+            },
+            multiple_lookups={
+                # unsigned
+                (LookupMultiple.parse_obj([2]), None): [2]
+            },
         )
 
         task = Sign(
             task_data={
                 "purpose": KeyPurpose.UEFI,
-                "unsigned": 2,
+                "unsigned": [2],
                 "key": key_lookup,
             }
         )
         self.assertEqual(
             task.compute_dynamic_data(task_db),
-            SignDynamicData(unsigned_id=2, key_id=1),
+            SignDynamicData(unsigned_ids=[2], key_id=1),
         )
 
     def test_fetch_input_missing_trusted_cert(self) -> None:
@@ -264,8 +265,8 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
             return self.make_artifact_response(signing_input, artifact_id)
 
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1},
-            dynamic_task_data={"unsigned_id": 2, "key_id": 1},
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1},
+            dynamic_task_data={"unsigned_ids": [2], "key_id": 1},
         )
         debusine_mock = self.mock_debusine(task)
         debusine_mock.download_artifact.side_effect = download_artifact
@@ -317,11 +318,11 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         task = Sign(
             task_data={
                 "purpose": KeyPurpose.UEFI,
-                "unsigned": signing_input_id,
+                "unsigned": [signing_input_id],
                 "key": signing_key_id,
             },
             dynamic_task_data={
-                "unsigned_id": signing_input_id,
+                "unsigned_ids": [signing_input_id],
                 "key_id": signing_key_id,
             },
         )
@@ -332,7 +333,7 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         with override_settings(DEBUSINE_SIGNING_TRUSTED_CERTS=[trusted_cert]):
             self.assertTrue(task.fetch_input(destination))
 
-        self.assertEqual(task._files, responses[2].files)
+        self.assertEqual(task._files, {signing_input_id: responses[2].files})
         self.assertEqual(task._key, key)
 
     def test_fetch_input_mismatched_purpose(self) -> None:
@@ -362,11 +363,11 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         task = Sign(
             task_data={
                 "purpose": KeyPurpose.OPENPGP,
-                "unsigned": signing_input_id,
+                "unsigned": [signing_input_id],
                 "key": signing_key_id,
             },
             dynamic_task_data={
-                "unsigned_id": signing_input_id,
+                "unsigned_ids": [signing_input_id],
                 "key_id": signing_key_id,
             },
         )
@@ -411,11 +412,11 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         task = Sign(
             task_data={
                 "purpose": KeyPurpose.UEFI,
-                "unsigned": signing_input_id,
+                "unsigned": [signing_input_id],
                 "key": signing_key_id,
             },
             dynamic_task_data={
-                "unsigned_id": signing_input_id,
+                "unsigned_ids": [signing_input_id],
                 "key_id": signing_key_id,
             },
         )
@@ -433,6 +434,65 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
             f"Signing key uefi:{fingerprint} does not exist\n",
         )
 
+    def test_fetch_input_multiple_unsigned_ids(self) -> None:
+        """fetch_input handles multiple unsigned artifacts."""
+        fingerprint = "0" * 64
+        key = Key.objects.create(
+            purpose=Key.Purpose.UEFI,
+            fingerprint=fingerprint,
+            private_key={},
+            public_key=b"",
+        )
+        signing_key = SigningKeyArtifact.create(
+            KeyPurpose.UEFI, fingerprint=fingerprint, public_key=b""
+        )
+        signing_key_id = 1
+        signing_inputs = [
+            SigningInputArtifact.create([], self.create_temporary_directory())
+            for _ in range(2)
+        ]
+        signing_input_ids = [2, 3]
+        responses = {
+            artifact_id: self.make_artifact_response(artifact, artifact_id)
+            for artifact, artifact_id in (
+                (signing_key, signing_key_id),
+                (signing_inputs[0], signing_input_ids[0]),
+                (signing_inputs[1], signing_input_ids[1]),
+            )
+        }
+
+        def download_artifact(
+            artifact_id: int, destination: Path, **kwargs: Any  # noqa: U100
+        ) -> ArtifactResponse:
+            return responses[artifact_id]
+
+        task = Sign(
+            task_data={
+                "purpose": KeyPurpose.UEFI,
+                "unsigned": signing_input_ids,
+                "key": signing_key_id,
+            },
+            dynamic_task_data={
+                "unsigned_ids": signing_input_ids,
+                "key_id": signing_key_id,
+            },
+        )
+        debusine_mock = self.mock_debusine(task)
+        debusine_mock.download_artifact.side_effect = download_artifact
+        destination = self.create_temporary_directory()
+
+        with override_settings(DEBUSINE_SIGNING_TRUSTED_CERTS=[]):
+            self.assertTrue(task.fetch_input(destination))
+
+        self.assertEqual(
+            task._files,
+            {
+                signing_input_id: responses[signing_input_id].files
+                for signing_input_id in signing_input_ids
+            },
+        )
+        self.assertEqual(task._key, key)
+
     def test_run_error(self) -> None:
         """run() records an error result if signing fails."""
         key = Key.objects.create(
@@ -445,10 +505,10 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         execute_directory = self.create_temporary_directory()
         (image := download_directory / "image").touch()
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         task.work_request_id = 1
-        task._files = self.make_files_response({"image": image})
+        task._files = {2: self.make_files_response({"image": image})}
         task._key = key
         debug_log_files_directory = TemporaryDirectory(prefix="debusine-tests-")
         self.addCleanup(debug_log_files_directory.cleanup)
@@ -456,10 +516,11 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         task.prepare_to_run(download_directory, execute_directory)
 
         with mock.patch.object(key, "sign", side_effect=Exception("Boom")):
-            self.assertTrue(task.run(execute_directory))
+            self.assertFalse(task.run(execute_directory))
 
         self.assertEqual(
-            task._results, [SigningResult(file="image", error_message="Boom")]
+            task._results,
+            {2: [SigningResult(file="image", error_message="Boom")]},
         )
 
     def test_run_success(self) -> None:
@@ -474,10 +535,10 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
         execute_directory = self.create_temporary_directory()
         (image := download_directory / "image").touch()
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         task.work_request_id = 1
-        task._files = self.make_files_response({"image": image})
+        task._files = {2: self.make_files_response({"image": image})}
         task._key = key
         debug_log_files_directory = TemporaryDirectory(prefix="debusine-tests-")
         self.addCleanup(debug_log_files_directory.cleanup)
@@ -489,15 +550,98 @@ class SignTests(SignTestMixin, TestHelpersMixin, TestCase):
 
         self.assertEqual(
             task._results,
-            [SigningResult(file="image", output_file="image.sig")],
+            {2: [SigningResult(file="image", output_file="image.sig")]},
         )
         mock_sign.assert_called_once_with(
             execute_directory / "input" / "image",
             execute_directory / "output" / "image.sig",
+            SigningMode.DETACHED,
             1,
             log_file=mock.ANY,
         )
 
+    def test_run_uefi_on_kmod(self) -> None:
+        """run() skips kernel modules for UEFI signing."""
+        key = Key.objects.create(
+            purpose=Key.Purpose.UEFI,
+            fingerprint="0" * 64,
+            private_key={},
+            public_key=b"",
+        )
+        download_directory = self.create_temporary_directory()
+        execute_directory = self.create_temporary_directory()
+        (image := download_directory / "image").touch()
+        (foo_ko := download_directory / "foo.ko").touch()
+        task = Sign(
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
+        )
+        task.work_request_id = 1
+        task._files = {
+            2: self.make_files_response({"image": image, "foo.ko": foo_ko})
+        }
+        task._key = key
+        debug_log_files_directory = TemporaryDirectory(prefix="debusine-tests-")
+        self.addCleanup(debug_log_files_directory.cleanup)
+        task._debug_log_files_directory = debug_log_files_directory
+        task.prepare_to_run(download_directory, execute_directory)
+
+        with mock.patch.object(key, "sign"):
+            self.assertTrue(task.run(execute_directory))
+
+        self.assertEqual(
+            task._results,
+            {
+                2: [
+                    SigningResult(file="image", output_file="image.sig"),
+                    SigningResult(
+                        file="foo.ko",
+                        error_message=(
+                            "Skipped kernel module (not yet supported)"
+                        ),
+                    ),
+                ]
+            },
+        )
+
+    def test_run_uefi_on_only_kmod(self) -> None:
+        """run() fails if no files are signed."""
+        key = Key.objects.create(
+            purpose=Key.Purpose.UEFI,
+            fingerprint="0" * 64,
+            private_key={},
+            public_key=b"",
+        )
+        download_directory = self.create_temporary_directory()
+        execute_directory = self.create_temporary_directory()
+        (foo_ko := download_directory / "foo.ko").touch()
+        task = Sign(
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
+        )
+        task.work_request_id = 1
+        task._files = {2: self.make_files_response({"foo.ko": foo_ko})}
+        task._key = key
+        debug_log_files_directory = TemporaryDirectory(prefix="debusine-tests-")
+        self.addCleanup(debug_log_files_directory.cleanup)
+        task._debug_log_files_directory = debug_log_files_directory
+        task.prepare_to_run(download_directory, execute_directory)
+
+        with mock.patch.object(key, "sign"):
+            self.assertFalse(task.run(execute_directory))
+
+        self.assertEqual(
+            task._results,
+            {
+                2: [
+                    SigningResult(
+                        file="foo.ko",
+                        error_message=(
+                            "Skipped kernel module (not yet supported)"
+                        ),
+                    ),
+                ]
+            },
+        )
+
 
 class SignIntegrationTests(
     SignTestMixin, TestHelpersMixin, TransactionTestCase
@@ -527,26 +671,34 @@ class SignIntegrationTests(
         (minimal_pecoff_path := temp_path / "image").write_bytes(
             minimal_pecoff_image
         )
-        signing_input = SigningInputArtifact.create(
-            [minimal_pecoff_path], temp_path
+        (minimal_pecoff_path2 := temp_path / "image2").write_bytes(
+            minimal_pecoff_image[:-1] + b"\0"
         )
-        signing_input_id = 2
+        signing_inputs = [
+            SigningInputArtifact.create([path], temp_path)
+            for path in (minimal_pecoff_path, minimal_pecoff_path2)
+        ]
+        signing_input_ids = [2, 3]
         responses = {
-            response.id: response
-            for response in (
-                self.make_artifact_response(signing_key, signing_key_id),
-                self.make_artifact_response(signing_input, signing_input_id),
+            artifact_id: self.make_artifact_response(artifact, artifact_id)
+            for artifact, artifact_id in (
+                (signing_key, signing_key_id),
+                (signing_inputs[0], signing_input_ids[0]),
+                (signing_inputs[1], signing_input_ids[1]),
             )
         }
         uploaded_paths: list[Path] = []
         output_path = self.create_temporary_directory()
-        signing_output_id = 3
+        signing_output_ids = [4, 5]
+        work_request_debug_logs_id = 6
 
         def download_artifact(
             artifact_id: int, destination: Path, **kwargs: Any  # noqa: U100
         ) -> ArtifactResponse:
-            if artifact_id == signing_input_id:
+            if artifact_id == signing_input_ids[0]:
                 shutil.copy(minimal_pecoff_path, destination)
+            elif artifact_id == signing_input_ids[1]:
+                shutil.copy(minimal_pecoff_path2, destination)
             return responses[artifact_id]
 
         def upload_artifact(
@@ -556,16 +708,25 @@ class SignIntegrationTests(
             for path in local_artifact.files.values():
                 uploaded_paths.append(path)
                 shutil.copy(path, output_path)
-            return RemoteArtifact(id=signing_output_id, workspace="System")
+            match (local_artifact.category, list(local_artifact.files)[0]):
+                case (ArtifactCategory.SIGNING_OUTPUT, "image.sig"):
+                    artifact_id = signing_output_ids[0]
+                case (ArtifactCategory.SIGNING_OUTPUT, "image2.sig"):
+                    artifact_id = signing_output_ids[1]
+                case (ArtifactCategory.WORK_REQUEST_DEBUG_LOGS, _):
+                    artifact_id = work_request_debug_logs_id
+                case _ as unreachable:
+                    raise AssertionError(f"unexpected upload: {unreachable}")
+            return RemoteArtifact(id=artifact_id, workspace="System")
 
         task = Sign(
             task_data={
                 "purpose": KeyPurpose.UEFI,
-                "unsigned": signing_input_id,
+                "unsigned": signing_input_ids,
                 "key": signing_key_id,
             },
             dynamic_task_data={
-                "unsigned_id": signing_input_id,
+                "unsigned_ids": signing_input_ids,
                 "key_id": signing_key_id,
             },
         )
@@ -584,7 +745,7 @@ class SignIntegrationTests(
             self.assertTrue(task.execute())
 
         # The task uploaded an artifact with the expected properties.
-        self.assertEqual(len(uploaded_paths), 2)
+        self.assertEqual(len(uploaded_paths), 3)
         debusine_mock.upload_artifact.assert_has_calls(
             [
                 mock.call(
@@ -604,6 +765,23 @@ class SignIntegrationTests(
                     workspace="System",
                     work_request=1,
                 ),
+                mock.call(
+                    SigningOutputArtifact(
+                        category=SigningOutputArtifact._category,
+                        files={"image2.sig": uploaded_paths[1]},
+                        data=DebusineSigningOutput(
+                            purpose=KeyPurpose.UEFI,
+                            fingerprint=key.fingerprint,
+                            results=[
+                                SigningResult(
+                                    file="image2", output_file="image2.sig"
+                                )
+                            ],
+                        ),
+                    ),
+                    workspace="System",
+                    work_request=1,
+                ),
                 mock.call(
                     WorkRequestDebugLogs(
                         category=WorkRequestDebugLogs._category,
@@ -622,18 +800,35 @@ class SignIntegrationTests(
         )
         debusine_mock.relation_create.assert_has_calls(
             [
-                mock.call(signing_output_id, signing_input_id, "relates-to"),
-                mock.call(signing_output_id, signing_key_id, "relates-to"),
+                mock.call(
+                    signing_output_ids[0], signing_input_ids[0], "relates-to"
+                ),
+                mock.call(signing_output_ids[0], signing_key_id, "relates-to"),
+                mock.call(
+                    signing_output_ids[1], signing_input_ids[1], "relates-to"
+                ),
+                mock.call(signing_output_ids[1], signing_key_id, "relates-to"),
             ]
         )
 
-        # The file was properly signed.
+        # The files were properly signed.
         (certificate := temp_path / "tmp.crt").write_bytes(key.public_key)
-        subprocess.run(
-            ["sbverify", "--cert", certificate, output_path / "image.sig"],
-            check=True,
-            stdout=subprocess.DEVNULL,
-        )
+        for image_path, signature_path in (
+            (minimal_pecoff_path, output_path / "image.sig"),
+            (minimal_pecoff_path2, output_path / "image2.sig"),
+        ):
+            subprocess.run(
+                [
+                    "sbverify",
+                    "--cert",
+                    certificate,
+                    "--detached",
+                    signature_path,
+                    image_path,
+                ],
+                check=True,
+                stdout=subprocess.DEVNULL,
+            )
 
     def test_openpgp(self) -> None:
         """Integration test: sign using a real OpenPGP key."""
@@ -689,11 +884,11 @@ class SignIntegrationTests(
         task = Sign(
             task_data={
                 "purpose": KeyPurpose.OPENPGP,
-                "unsigned": signing_input_id,
+                "unsigned": [signing_input_id],
                 "key": signing_key_id,
             },
             dynamic_task_data={
-                "unsigned_id": signing_input_id,
+                "unsigned_ids": [signing_input_id],
                 "key_id": signing_key_id,
             },
         )
@@ -777,6 +972,6 @@ class SignIntegrationTests(
     def test_label(self) -> None:
         """Test get_label."""
         task = Sign(
-            task_data={"purpose": KeyPurpose.UEFI, "unsigned": 2, "key": 1}
+            task_data={"purpose": KeyPurpose.UEFI, "unsigned": [2], "key": 1}
         )
         self.assertEqual(task.get_label(), "sign data for uefi")
