@@ -14,14 +14,10 @@ from pathlib import Path
 from typing import Any
 
 from debian import debfile
-
 from debusine import utils
 from debusine.artifacts import LintianArtifact
 from debusine.artifacts.models import CollectionCategory, DebianLintianSummary
-from debusine.tasks import (
-    BaseTaskWithExecutor,
-    RunCommandTask,
-)
+from debusine.tasks import BaseTaskWithExecutor, RunCommandTask
 from debusine.tasks.models import (
     LintianData,
     LintianDynamicData,
@@ -93,22 +89,18 @@ class Lintian(
         self, task_database: TaskDatabaseInterface
     ) -> LintianDynamicData:
         """Resolve artifact lookups for this task."""
-        environment = self.get_environment_lookup()
         return LintianDynamicData(
             environment_id=(
                 None
-                if environment is None
-                else task_database.lookup_single_artifact(
-                    environment,
+                if self.data.environment is None
+                else self.get_environment(
+                    task_database,
+                    self.data.environment,
                     default_category=CollectionCategory.ENVIRONMENTS,
                 )
             ),
-            input_source_artifact_id=(
-                None
-                if self.data.input.source_artifact is None
-                else task_database.lookup_single_artifact(
-                    self.data.input.source_artifact
-                )
+            input_source_artifact_id=task_database.lookup_single_artifact(
+                self.data.input.source_artifact
             ),
             input_binary_artifacts_ids=task_database.lookup_multiple_artifacts(
                 self.data.input.binary_artifacts
@@ -476,7 +468,7 @@ class Lintian(
 
     def upload_artifacts(
         self, exec_directory: Path, *, execution_success: bool  # noqa: U100
-    ):
+    ) -> None:
         """Upload the LintianArtifact with the files and relationships."""
         if not self.debusine:
             raise AssertionError("self.debusine not set")
