@@ -9,9 +9,11 @@
 
 """Models used by debusine tasks."""
 
+import re
 from datetime import datetime
 from enum import Enum, StrEnum, auto
 from itertools import groupby
+from pathlib import Path
 from typing import Annotated, Any, Literal, TypeAlias, Union
 
 try:
@@ -22,6 +24,7 @@ except ImportError:
 import yaml
 
 from debusine.artifacts.models import BareDataCategory
+from debusine.client.models import LookupChildType
 
 
 # taken from functional.py in python3-django 3:4.2.11-1 to avoid the dependency
@@ -50,6 +53,7 @@ class TaskTypes(StrEnum):
     INTERNAL = "Internal"
     WORKFLOW = "Workflow"
     SIGNING = "Signing"
+    WAIT = "Wait"
 
     @classproperty
     def choices(cls):
@@ -157,14 +161,6 @@ def parse_key_value_lookup_segment(segment: str) -> tuple[str, dict[str, str]]:
     return lookup_type, dict(item.split("=", 1) for item in filters_segments)
 
 
-class LookupChildType(StrEnum):
-    """Possible values for LookupDict.child_type."""
-
-    BARE = "bare"
-    ARTIFACT = "artifact"
-    COLLECTION = "collection"
-
-
 class CollectionItemMatcherKind(StrEnum):
     """Possible values for CollectionItemMatcher.kind."""
 
@@ -194,6 +190,13 @@ class LookupDict(BaseTaskDataModel):
     # Logically a mapping, but we turn it into a tuple so that lookups are
     # hashable.
     data_matchers: tuple[tuple[str, CollectionItemMatcher], ...] = ()
+    # Logically a mapping, but we turn it into a tuple so that lookups are
+    # hashable.  In future some filters may want other value types, but they
+    # need to be hashable and our current use cases involve subordinate
+    # lookups, so just declare them to be lookups for now.
+    lookup_filters: tuple[
+        tuple[str, Union[LookupSingle, "LookupMultiple"]], ...
+    ] = ()
 
     @pydantic.root_validator(pre=True, allow_reuse=True)
     @classmethod
@@ -203,8 +206,9 @@ class LookupDict(BaseTaskDataModel):
 
         `name`, `name__*`, `data__KEY`, and `data__KEY__*` keys in `values`
         are transformed into :py:class:`CollectionItemMatcher` instances
-        stored in `name_matcher` and `data_matchers`.  Conflicting lookup
-        suffixes are rejected.
+        stored in `name_matcher` and `data_matchers`.  `lookup__KEY` keys
+        are transformed into entries in `lookup_filters`.  Conflicting
+        lookup suffixes are rejected.
         """
 
         def split_words(s: str) -> list[str]:
@@ -216,19 +220,24 @@ class LookupDict(BaseTaskDataModel):
                     return ("name",)
                 case ["data", data_key, *_]:
                     return ("data", data_key)
+                case ["lookup", lookup_key]:
+                    return ("lookup", lookup_key)
                 case _:
                     return ()
 
         data_matchers: dict[str, CollectionItemMatcher] = dict(
             values.get("data_matchers", ())
         )
+        lookup_filters: dict[str, LookupSingle | LookupMultiple] = dict(
+            values.get("lookup_filters", ())
+        )
 
         for key, group in groupby(
             sorted(values, key=matcher_prefix), matcher_prefix
         ):
             if not key:
-                # Only keys with `name` or `data` as their first segment
-                # need transformation.
+                # Only keys with `name`, `data`, or `lookup` as their first
+                # segment need transformation.
                 continue
             matcher_names = list(group)
             if len(matcher_names) > 1:
@@ -248,10 +257,18 @@ class LookupDict(BaseTaskDataModel):
                     kind = "exact"
                 case ["data", data_key, kind] if kind != "exact":
                     pass
+                case ["lookup", lookup_key]:
+                    lookup_filters[lookup_key] = pydantic.parse_obj_as(
+                        LookupSingle | LookupMultiple,  # type: ignore[arg-type]
+                        values[matcher_name],
+                    )
+                    for k in matcher_names:
+                        del values[k]
+                    continue
                 case _:
                     # We filter out keys with first segments other than
-                    # `name` and `data` earlier, so anything that reaches
-                    # here must be a malformed lookup.
+                    # `name`, `data`, or `lookup` earlier, so anything that
+                    # reaches here must be a malformed lookup.
                     raise ValueError(f"Unrecognized matcher: {matcher_name}")
 
             matcher = CollectionItemMatcher(
@@ -268,6 +285,8 @@ class LookupDict(BaseTaskDataModel):
 
         if data_matchers:
             values["data_matchers"] = tuple(sorted(data_matchers.items()))
+        if lookup_filters:
+            values["lookup_filters"] = tuple(sorted(lookup_filters.items()))
 
         return values
 
@@ -278,7 +297,7 @@ class LookupDict(BaseTaskDataModel):
         This reverses the transformations applied by
         :py:meth:`normalize_matchers`.
         """
-        value = {"collection": self.collection}
+        value: dict[str, Any] = {"collection": self.collection}
         if self.child_type != LookupChildType.ARTIFACT:
             value["child_type"] = str(self.child_type)
         if self.category is not None:
@@ -295,6 +314,11 @@ class LookupDict(BaseTaskDataModel):
                 value[f"data__{key}"] = matcher.value
             else:
                 value[f"data__{key}__{matcher.kind}"] = matcher.value
+        for key, filter_value in self.lookup_filters:
+            if isinstance(filter_value, LookupMultiple):
+                value[f"lookup__{key}"] = filter_value.export()
+            else:
+                value[f"lookup__{key}"] = filter_value
         return value
 
 
@@ -341,6 +365,10 @@ class LookupMultiple(BaseTaskDataModel):
             ]
 
 
+# Resolve circular reference to LookupMultiple.
+LookupDict.update_forward_refs()
+
+
 class NotificationDataEmail(BaseTaskDataModel):
     """Channel data for email notifications."""
 
@@ -356,6 +384,7 @@ class ActionTypes(StrEnum):
     SEND_NOTIFICATION = "send-notification"
     UPDATE_COLLECTION_WITH_ARTIFACTS = "update-collection-with-artifacts"
     UPDATE_COLLECTION_WITH_DATA = "update-collection-with-data"
+    RETRY_WITH_DELAYS = "retry-with-delays"
 
 
 class ActionSendNotification(BaseTaskDataModel):
@@ -392,11 +421,35 @@ class ActionUpdateCollectionWithData(BaseTaskDataModel):
     data: dict[str, Any] | None = None
 
 
+class ActionRetryWithDelays(BaseTaskDataModel):
+    """Action for retrying a work request with delays."""
+
+    _delay_re = re.compile(r"^([0-9]+)([mhdw])$")
+
+    action: Literal[ActionTypes.RETRY_WITH_DELAYS] = (
+        ActionTypes.RETRY_WITH_DELAYS
+    )
+    delays: list[str] = pydantic.Field(min_items=1)
+
+    @pydantic.validator("delays")
+    @classmethod
+    def validate_delays(cls, values: list[str]) -> list[str]:
+        """Check items in `delays` field."""
+        for v in values:
+            if cls._delay_re.match(v) is None:
+                raise ValueError(
+                    f"Item in delays must be an integer followed by "
+                    f"m, h, d, or w; got {v!r}"
+                )
+        return values
+
+
 EventReaction = Annotated[
     Union[
         ActionSendNotification,
         ActionUpdateCollectionWithArtifacts,
         ActionUpdateCollectionWithData,
+        ActionRetryWithDelays,
     ],
     pydantic.Field(discriminator="action"),
 ]
@@ -500,7 +553,7 @@ class AutopkgtestData(BaseTaskDataWithExecutor):
     fail_on: AutopkgtestFailOn = pydantic.Field(
         default_factory=AutopkgtestFailOn
     )
-    timeout: AutopkgtestTimeout | None
+    timeout: AutopkgtestTimeout | None = None
     # BaseTaskDataWithExecutor declares this as optional, but it's required
     # here.
     environment: LookupSingle
@@ -578,12 +631,6 @@ class BlhcInput(BaseTaskDataModel):
     artifact: LookupSingle
 
 
-class BlhcOutput(BaseTaskDataModel):
-    """Output configuration for a blhc task."""
-
-    source_analysis: bool = True
-
-
 class BlhcFlags(StrEnum):
     """Possible values for extra_flags."""
 
@@ -600,7 +647,6 @@ class BlhcData(BaseTaskDataWithExecutor):
     """In memory task data for the Blhc task."""
 
     input: BlhcInput
-    output: BlhcOutput = pydantic.Field(default_factory=BlhcOutput)
     # Passed to blhc
     extra_flags: list[BlhcFlags] = pydantic.Field(default_factory=list)
 
@@ -611,6 +657,69 @@ class BlhcDynamicData(BaseDynamicTaskDataWithExecutor):
     input_artifact_id: int
 
 
+class DebDiffInput(BaseTaskDataModel):
+    """Input for a debdiff task."""
+
+    source_artifacts: list[LookupSingle] | None = pydantic.Field(
+        None, min_items=2, max_items=2
+    )
+    binary_artifacts: list[LookupMultiple] | None = pydantic.Field(
+        None, min_items=2, max_items=2
+    )
+
+    @pydantic.root_validator(allow_reuse=True)
+    @classmethod
+    def check_one_of_source_or_binary(cls, values):
+        """Ensure only source or binary artifacts are present."""
+        if (values.get("source_artifacts") is None) == (
+            values.get("binary_artifacts") is None
+        ):
+            raise ValueError(
+                'Exactly one of source_artifacts'
+                'or binary_artifacts must be set'
+            )
+        return values
+
+
+class DebDiffFlags(StrEnum):
+    """Possible values for extra_flags."""
+
+    DIRS = "--dirs"
+    NOCONTROL = "--nocontrol"
+    WDIFF = "--wdiff-source-control"
+    SHOW_MOVED = "--show-moved"
+    DIFFSTAT = "--diffstat"
+    PATCHES = "--apply-patches"
+    IGNORE_SPACE = "--ignore-space"
+
+
+class DebDiffData(BaseTaskDataWithExecutor):
+    """In memory task data for the DebDiff task."""
+
+    input: DebDiffInput
+    # Passed to debdiff
+    extra_flags: list[DebDiffFlags] = pydantic.Field(default_factory=list)
+    host_architecture: str
+    # BaseTaskDataWithExecutor declares this as optional, but it's required
+    # here.
+    environment: LookupSingle
+
+
+class DebDiffDynamicData(BaseDynamicTaskDataWithExecutor):
+    """Dynamic data for the DebDiff task."""
+
+    # BaseTaskDynamicDataWithExecutor declares this as optional, but it's
+    # required here.
+    environment_id: int
+
+    input_source_artifacts_ids: list[int] | None = pydantic.Field(
+        None, min_items=2, max_items=2
+    )
+    # TODO: Once we depend on pydantic 2.9.2-1 add a check:
+    # pydantic.Field(None, min_items=2, max_items=2)
+    input_binary_artifacts_ids: list[list[int]] | None = None
+
+
 class MmDebstrapVariant(StrEnum):
     """Possible values for variant."""
 
@@ -660,10 +769,32 @@ class SystemBootstrapRepositoryCheckSignatureWith(StrEnum):
 class SystemBootstrapRepositoryKeyring(BaseTaskDataModel):
     """Description of a repository keyring."""
 
-    url: pydantic.AnyUrl
+    url: pydantic.AnyUrl | pydantic.FileUrl
     sha256sum: str = ""
     install: bool = False
 
+    @pydantic.validator("url")
+    @classmethod
+    def validate_url(
+        cls, url: pydantic.AnyUrl | pydantic.FileUrl
+    ) -> pydantic.AnyUrl | pydantic.FileUrl:
+        """
+        Reject file:// URLs outside /usr/share/keyrings/.
+
+        We don't want to allow reading arbitrary paths.
+        """
+        if (
+            url.scheme == "file"
+            and url.path is not None
+            and not (
+                Path(url.path).resolve().is_relative_to("/usr/share/keyrings")
+            )
+        ):
+            raise ValueError(
+                "file:// URLs for keyrings must be under /usr/share/keyrings/"
+            )
+        return url
+
 
 class SystemBootstrapRepository(BaseTaskDataModel):
     """Description of one repository in SystemBootstrapData."""
@@ -831,6 +962,7 @@ class SbuildData(BaseTaskDataWithExecutor):
         default_factory=lambda: [SbuildBuildComponent.ANY],
     )
     binnmu: SbuildBinNMU | None = None
+    build_profiles: list[str] | None
 
     @pydantic.root_validator
     @classmethod
@@ -858,6 +990,17 @@ class SbuildData(BaseTaskDataWithExecutor):
                 )
         return values
 
+    @pydantic.root_validator
+    @classmethod
+    def check_binnmu_against_components(cls, values):
+        """Binnmus are incompatible with architecture-independent builds."""
+        if values.get("binnmu") is not None:
+            if SbuildBuildComponent.ALL in values.get("build_components"):
+                raise ValueError(
+                    "Cannot build architecture-independent packages in a binNMU"
+                )
+        return values
+
 
 class SbuildDynamicData(BaseDynamicTaskDataWithExecutor):
     """Dynamic data for the Sbuild task."""
@@ -866,6 +1009,7 @@ class SbuildDynamicData(BaseDynamicTaskDataWithExecutor):
     input_extra_binary_artifacts_ids: list[int] = pydantic.Field(
         default_factory=list
     )
+    binnmu_maintainer: str | None = None
 
 
 class ImageCacheUsageLogEntry(BaseTaskDataModel):
@@ -973,6 +1117,60 @@ class AssembleSignedSourceDynamicData(BaseDynamicTaskDataWithExecutor):
     signed_ids: list[int]
 
 
+class MakeSourcePackageUploadInput(BaseTaskDataModel):
+    """Input for a MakeSourcePackageUpload task."""
+
+    source_artifact: LookupSingle
+
+
+class MakeSourcePackageUploadData(BaseTaskDataWithExecutor):
+    """In memory task data for the MakeSourcePackageUpload task."""
+
+    # BaseTaskDataWithExecutor declares this as optional, but it's required
+    # here.
+    environment: LookupSingle
+
+    input: MakeSourcePackageUploadInput
+    since_version: str | None
+    target_distribution: str | None
+
+
+class MakeSourcePackageUploadDynamicData(BaseDynamicTaskDataWithExecutor):
+    """Expanded dynamic data for the MakeSourcePackageUpload task."""
+
+    # BaseDynamicTaskDataWithExecutor declares this as optional, but it's
+    # required here.
+    environment_id: int
+
+    input_source_artifact_id: int
+
+
+class MergeUploadsInput(BaseTaskDataModel):
+    """Input for a MergeUploads task."""
+
+    uploads: LookupMultiple
+
+
+class MergeUploadsData(BaseTaskDataWithExecutor):
+    """In memory task data for the MergeUploads task."""
+
+    # BaseTaskDataWithExecutor declares this as optional, but it's required
+    # here.
+    environment: LookupSingle
+
+    input: MergeUploadsInput
+
+
+class MergeUploadsDynamicData(BaseDynamicTaskDataWithExecutor):
+    """Expanded dynamic data for the MergeUploads task."""
+
+    # BaseDynamicTaskDataWithExecutor declares this as optional, but it's
+    # required here.
+    environment_id: int
+
+    input_uploads_ids: list[int]
+
+
 # Workarounds for https://github.com/yaml/pyyaml/issues/722
 yaml.SafeDumper.add_multi_representer(
     StrEnum,
