@@ -15,17 +15,17 @@ https://freexian-team.pages.debian.net/debusine/reference/tasks.html#task-packag
 """
 
 import email.utils
+import os
 import subprocess
 from pathlib import Path
 from typing import Any, cast
 
 try:
-    from pydantic.v1 import ValidationError
+    import pydantic.v1 as pydantic
 except ImportError:
-    from pydantic import ValidationError  # type: ignore
+    import pydantic  # type: ignore
 
 import debian.deb822 as deb822
-
 import yaml
 
 import debusine.utils
@@ -42,11 +42,7 @@ from debusine.artifacts.models import (
     DoseDistCheck,
 )
 from debusine.client.models import RemoteArtifact
-from debusine.tasks import (
-    BaseTaskWithExecutor,
-    RunCommandTask,
-    TaskConfigError,
-)
+from debusine.tasks import BaseTaskWithExecutor, RunCommandTask, TaskConfigError
 from debusine.tasks.models import SbuildData, SbuildDynamicData
 from debusine.tasks.sbuild_validator_mixin import SbuildValidatorMixin
 from debusine.tasks.server import TaskDatabaseInterface
@@ -80,13 +76,14 @@ class Sbuild(
         self, task_database: TaskDatabaseInterface
     ) -> SbuildDynamicData:
         """Resolve artifact lookups for this task."""
-        environment = self.get_environment_lookup()
+        DEBUSINE_FQDN = task_database.get_server_setting("DEBUSINE_FQDN")
         return SbuildDynamicData(
             environment_id=(
                 None
-                if environment is None
-                else task_database.lookup_single_artifact(
-                    environment,
+                if self.data.environment is None
+                else self.get_environment(
+                    task_database,
+                    self.data.environment,
                     default_category=CollectionCategory.ENVIRONMENTS,
                 )
             ),
@@ -98,6 +95,7 @@ class Sbuild(
                     self.data.input.extra_binary_artifacts
                 )
             ),
+            binnmu_maintainer=f"Debusine <noreply@{DEBUSINE_FQDN}>",
         )
 
     def get_source_artifacts_ids(self) -> list[int]:
@@ -144,9 +142,6 @@ class Sbuild(
             chroots_key = cls.prefix_with_task_name("chroots")
             metadata[chroots_key] = cls._list_chroots()
 
-        host_arch_key = cls.prefix_with_task_name("host_architecture")
-        metadata[host_arch_key] = cls._call_dpkg_architecture()
-
         return metadata
 
     def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
@@ -334,8 +329,16 @@ class Sbuild(
                     self.data.binnmu.timestamp
                 )
                 cmd.append(f"--binNMU-timestamp={rfc5322date}")
+            maintainer: str | None
             if self.data.binnmu.maintainer is not None:
-                cmd.append(f"--maintainer={self.data.binnmu.maintainer}")
+                maintainer = str(self.data.binnmu.maintainer)
+            else:
+                assert self.dynamic_data
+                maintainer = self.dynamic_data.binnmu_maintainer
+            cmd.append(f"--maintainer={maintainer}")
+
+        if self.data.build_profiles:
+            cmd.append(f"--profiles={','.join(self.data.build_profiles)}")
 
         # BD-Uninstallable
         cmd.append("--bd-uninstallable-explainer=dose3")
@@ -347,6 +350,21 @@ class Sbuild(
 
         return cmd
 
+    def _cmd_env(self) -> dict[str, str] | None:
+        """Set DEB_BUILD_OPTIONS for build profiles."""
+        build_options = [
+            profile
+            for profile in self.data.build_profiles or []
+            if profile in ("nocheck", "nodoc")
+        ]
+        if build_options:
+            env = dict(os.environ)
+            env["DEB_BUILD_OPTIONS"] = " ".join(
+                os.environ.get("DEB_BUILD_OPTIONS", "").split() + build_options
+            )
+            return env
+        return None
+
     def execute(self) -> bool:
         """
         Verify task can be executed and super().execute().
@@ -391,7 +409,7 @@ class Sbuild(
             parsed_dict = yaml.load(output, yaml.CBaseLoader)
             try:
                 return DoseDistCheck.parse_obj(parsed_dict)
-            except ValidationError:
+            except pydantic.ValidationError:
                 return None
 
     def _upload_package_build_log(
@@ -523,7 +541,7 @@ class Sbuild(
                 self._create_binary_package_local_artifacts(
                     build_directory,
                     dsc,
-                    host_arch,
+                    "all",
                     [prefix + ".deb", prefix + ".udeb"],
                 )
             )
@@ -604,7 +622,9 @@ class Sbuild(
 
         return True
 
-    def upload_artifacts(self, directory: Path, *, execution_success: bool):
+    def upload_artifacts(
+        self, directory: Path, *, execution_success: bool
+    ) -> None:
         """
         Upload the artifacts from directory.
 
