@@ -11,24 +11,23 @@ from django.urls import include, path
 
 from debusine.web.views.task_status import TaskStatusView
 
+# Scoped URLs
 urlpatterns = [
-    path("", include("debusine.web.urls.homepage", namespace="homepage")),
+    # TODO: move to unscoped
     path(
         "accounts/",
         include("debusine.web.urls.accounts", namespace="accounts"),
     ),
+    # TODO: move to unscoped, or filter by scope?
     path(
         "work-request/",
         include("debusine.web.urls.work_requests", namespace="work-requests"),
     ),
+    # TODO: move to unscoped
     path(
         "workers/",
         include("debusine.web.urls.workers", namespace="workers"),
     ),
-    path(
-        "workspace/",
-        include("debusine.web.urls.workspaces", namespace="workspaces"),
-    ),
     path(
         "artifact/",
         include(
@@ -36,6 +35,7 @@ urlpatterns = [
             namespace="artifacts",
         ),
     ),
+    # TODO: move to unscoped
     path(
         "user/",
         include(
@@ -43,9 +43,14 @@ urlpatterns = [
             namespace="user",
         ),
     ),
+    # TODO: move to unscoped, or filter by scope?
     path(
         "task-status/",
         TaskStatusView.as_view(),
         name="task-status",
     ),
+    path(
+        "",
+        include("debusine.web.urls.workspaces", namespace="workspaces"),
+    ),
 ]
