@@ -8,14 +8,12 @@
 # contained in the LICENSE file.
 
 """Tests for the view tests utilities."""
-from django.test import TestCase
-from django.urls import reverse
-
 import lxml
 import lxml.objectify
-
+from django.urls import reverse
 from rest_framework import status
 
+from debusine.test.django import TestCase
 from debusine.web.views.tests.utils import ViewTestMixin
 
 
@@ -42,6 +40,10 @@ class TestUtils(ViewTestMixin, TestCase):
 
     def test_workspace_list_not_found(self):
         """Test that workspace_list_table_rows errors when not found."""
+        default_workspace = self.playground.get_default_workspace()
+        default_workspace.public = False
+        default_workspace.save()
+
         response = self.client.get(reverse("homepage:homepage"))
         self.assertEqual(response.status_code, status.HTTP_200_OK)
         tree = self.assertHTMLValid(response)
