@@ -13,21 +13,22 @@ from datetime import timedelta
 from typing import ClassVar
 from unittest import mock
 
+from django.conf import settings
 from django.contrib import messages
 from django.contrib.auth import get_user_model
 from django.test import override_settings
 from django.urls import reverse
-
 from rest_framework import status
 
 from debusine.db.models import (
+    Scope,
     Token,
     User,
     WorkRequest,
     WorkflowTemplate,
     default_workspace,
 )
-from debusine.test.django import TestCase
+from debusine.test.django import ListFilter, TestCase, override_permission
 from debusine.web.templatetags import debusine as template_debusine
 from debusine.web.views import UserTokenDeleteView
 from debusine.web.views.base import Widget
@@ -61,14 +62,6 @@ class HomepageViewTests(ViewTestMixin, TestCase):
         "Create artifact</a>"
     )
 
-    def _get_logged_in_user_element(self, username: str) -> str:
-        return (
-            '<a class="nav-link dropdown-toggle" href="#" role="button"'
-            ' data-bs-toggle="dropdown" aria-expanded="false">'
-            + username
-            + '</a>'
-        )
-
     def test_html_check_icon(self) -> None:
         """Test that html_check_icon returns the right results."""
         self.assertIn("green", html_check_icon(True))
@@ -86,19 +79,27 @@ class HomepageViewTests(ViewTestMixin, TestCase):
 
     def test_homepage(self):
         """Homepage view loads."""
-        response = self.client.get(reverse("homepage:homepage"))
+        # Mark the default workspace as private in order to test the "No
+        # workspaces" message.
+        default_workspace = self.playground.get_default_workspace()
+        default_workspace.public = False
+        default_workspace.save()
 
+        response = self.client.get(reverse("homepage:homepage"))
         self.assertEqual(response.status_code, status.HTTP_200_OK)
-        self.assertContains(response, "Welcome to debusine!")
         tree = self.assertHTMLValid(response)
+
+        self.assertTextContentEqual(
+            tree.xpath("//h1")[0], "Welcome to debusine!"
+        )
+
         ul = tree.xpath("//ul[@id='navbar-right']")[0]
         self.assertEqual(
-            ul.li.a.get("href"), reverse("accounts:login") + "?next=/"
+            ul.li.a.get("href"),
+            reverse("accounts:login") + f"?next={reverse('homepage:homepage')}",
         )
 
-        self.assertNotContains(
-            response, self._get_logged_in_user_element("testuser"), html=True
-        )
+        self.assertFalse(tree.xpath("//ul[@id='navbar-user-actions']"))
 
         self.assertNotContains(response, self.list_your_tokens_html, html=True)
         self.assertNotContains(response, self.create_work_request_html)
@@ -109,6 +110,10 @@ class HomepageViewTests(ViewTestMixin, TestCase):
             response, self.no_workspaces_not_authenticated, html=True
         )
 
+        ul = tree.xpath("//ul[@id='scope-list']")[0]
+        self.assertEqual(ul.li.a.get("href"), reverse("workspaces:list"))
+        self.assertTextContentEqual(ul.li.a, "debusine")
+
     def test_homepage_logged_in(self):
         """User is logged in: contains "You are authenticated as: username"."""
         username = "testuser"
@@ -123,9 +128,8 @@ class HomepageViewTests(ViewTestMixin, TestCase):
 
         ul = tree.xpath("//ul[@id='navbar-right']")[0]
         self.assertTextContentEqual(ul.li.a, username)
-        self.assertEqual(
-            ul.li.ul.li[-1].a.get("href"), reverse("accounts:logout")
-        )
+        ul = tree.xpath("//ul[@id='navbar-user-actions']")[0]
+        self.assertEqual(ul.li[-1].a.get("href"), reverse("accounts:logout"))
 
         self.assertContains(response, self.list_your_tokens_html, html=True)
         self.assertContains(response, self.create_work_request_html, html=True)
@@ -154,10 +158,10 @@ class HomepageViewTests(ViewTestMixin, TestCase):
 
         self.client.force_login(user)
 
-        work_request1 = self.create_work_request(
+        work_request1 = self.playground.create_work_request(
             task_name="noop", created_by=user, id=11
         )
-        work_request2 = self.create_work_request(
+        work_request2 = self.playground.create_work_request(
             task_name="noop", created_by=user, id=12
         )
 
@@ -223,6 +227,24 @@ class HomepageViewTests(ViewTestMixin, TestCase):
         self.assertEqual(msgdiv.get("role"), "alert")
         self.assertEqual(msgdiv.div[1].text.strip(), "Error message")
 
+    def test_homepage_scope_list(self):
+        """Homepage lists visible scopes."""
+        scope1 = self.playground.get_or_create_scope("scope1")
+        self.playground.get_or_create_scope("scope2")
+
+        with override_permission(
+            Scope, "can_display", ListFilter, exclude=[scope1]
+        ):
+            response = self.client.get(reverse("homepage:homepage"))
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+        tree = self.assertHTMLValid(response)
+
+        ul = tree.xpath("//ul[@id='scope-list']")[0]
+        self.assertEqual(
+            [self.get_node_text_normalized(li) for li in ul.li],
+            ["debusine", "scope2"],
+        )
+
 
 class LoginViewTests(TestCase):
     """Tests for the LoginView class."""
@@ -680,6 +702,79 @@ class WidgetTests(TestCase):
             self.assertEqual(str(e.exception), "ERROR")
 
 
+class LegacyRedirectTests(TestCase):
+    """Test the best effort redirects for unscoped URLs."""
+
+    def test_scope_redirect_valid(self):
+        """Test redirection of a valid legacy URL."""
+        response = self.client.get("/accounts/foo/bar?baz=true")
+        self.assertEqual(
+            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
+        )
+        self.assertEqual(
+            response.headers["Location"],
+            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/accounts/foo/bar?baz=true",
+        )
+
+    def test_scope_redirect_invalid(self):
+        """Test redirection of an invalid legacy URL."""
+        # Misspelled url (misses trailing 's')
+        response = self.client.get("/account/foo/bar?baz=true")
+        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
+
+    def test_scope_reproduce_redirect_loop(self):
+        """Reproduce a redirect loop."""
+        artifact, _ = self.playground.create_artifact()
+
+        response = self.client.get(f"/api/1.0/artifact/{artifact.pk}")
+        self.assertEqual(
+            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
+        )
+        self.assertEqual(
+            response.headers["Location"], f"/api/1.0/artifact/{artifact.pk}/"
+        )
+
+        response = self.client.get(
+            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/artifact/{artifact.pk}"
+        )
+        self.assertEqual(
+            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
+        )
+        self.assertEqual(
+            response.headers["Location"],
+            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/artifact/{artifact.pk}/",
+        )
+
+    # urlconfs are created at Django startup, so changing the setting
+    # afterwards has no effect: the redirect replacement pattern has already
+    # been created. This test would be pointless, but keeping it here as
+    # documentation. I would not subclass RedirectView only to have a more
+    # dynamic replacement pattern for use only in tests.
+    #
+    # @override_settings(DEBUSINE_DEFAULT_SCOPE="debian")
+    # def test_redirect_different_scope(self):
+    #     """Redirection uses DEBUSINE_DEFAULT_SCOPE."""
+    #     response = self.client.get("/accounts/foo/bar?baz=true")
+    #     self.assertEqual(
+    #         response.status_code, status.HTTP_301_MOVED_PERMANENTLY
+    #     )
+    #     self.assertEqual(
+    #         response.headers["Location"],
+    #         "/debian/accounts/foo/bar?baz=true",
+    #     )
+
+    def test_workspace_redirect_valid(self):
+        """Test redirection of a valid legacy workspace URL."""
+        response = self.client.get("/debusine/workspace/foo/bar?baz=true")
+        self.assertEqual(
+            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
+        )
+        self.assertEqual(
+            response.headers["Location"],
+            "/debusine/foo/bar?baz=true",
+        )
+
+
 _sort_table_handle = '<span class="order-active">&#9660;</span>'
 
 
