@@ -14,6 +14,7 @@ from datetime import timedelta
 from types import GeneratorType
 from typing import ClassVar
 
+import yaml
 from django.contrib.auth import get_user_model
 from django.contrib.messages import constants as messages_constants
 from django.contrib.messages import get_messages
@@ -22,37 +23,33 @@ from django.core.exceptions import ValidationError
 from django.core.validators import ProhibitNullCharactersValidator
 from django.urls import reverse
 from django.utils import timezone
-
 from lxml import html
-
 from rest_framework import status
 
-import yaml
-
 from debusine.artifacts import LintianArtifact
+from debusine.db.context import context
 from debusine.db.models import (
     Artifact,
+    TaskDatabase,
     User,
     WorkRequest,
     WorkflowTemplate,
     Workspace,
 )
-from debusine.server.scheduler import TaskDatabase
 from debusine.server.workflows.models import (
     WorkRequestManualUnblockAction,
     WorkRequestManualUnblockData,
     WorkRequestManualUnblockLog,
 )
 from debusine.tasks import Lintian
+from debusine.tasks.models import TaskTypes
 from debusine.tasks.tests.test_simplesystemimagebuild import (
     SimpleSystemImageBuildTests,
 )
 from debusine.test.django import TestCase
 from debusine.web.views import ui_shortcuts
 from debusine.web.views.lintian import LintianView
-from debusine.web.views.tests.test_views import (
-    _sort_table_handle,
-)
+from debusine.web.views.tests.test_views import _sort_table_handle
 from debusine.web.views.tests.utils import ViewTestMixin
 from debusine.web.views.work_request import (
     WorkRequestDetailView,
@@ -68,23 +65,31 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
     work_request: ClassVar[WorkRequest]
 
     @classmethod
+    @context.disable_permission_checks()
     def setUpTestData(cls):
         """Set up common objects."""
         super().setUpTestData()
-        cls.workspace = cls.playground.get_default_workspace()
+        cls.workspace = cls.playground.create_workspace(
+            name="Public", public=True
+        )
         started_at = timezone.now()
         duration = 73
         completed_at = started_at + timedelta(seconds=duration)
 
-        environment_item = cls.playground.create_debian_environment()
+        environment_item = cls.playground.create_debian_environment(
+            workspace=cls.workspace
+        )
         assert environment_item.artifact is not None
         environment = environment_item.artifact
 
-        cls.source = cls.playground.create_source_artifact()
+        cls.source = cls.playground.create_source_artifact(
+            workspace=cls.workspace
+        )
         cls.work_request = cls.playground.create_sbuild_work_request(
             source=cls.source,
             architecture="all",
             environment=environment,
+            workspace=cls.workspace,
         )
         cls.playground.compute_dynamic_data(cls.work_request)
         cls.work_request.mark_running()
@@ -142,6 +147,36 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
             response, "<h2>Workflow information</h2>", html=True
         )
 
+    def test_get_work_request_private_workspace_unauthenticated(self):
+        """Work requests in private workspaces 403 to the unauthenticated."""
+        with context.disable_permission_checks():
+            private_workspace = self.playground.create_workspace(name="Private")
+        self.work_request.workspace = private_workspace
+        self.work_request.save()
+
+        response = self.client.get(
+            reverse("work-requests:detail", kwargs={"pk": self.work_request.id})
+        )
+        self.assertContains(
+            response,
+            "You need to be authenticated to view this Work Request",
+            status_code=status.HTTP_403_FORBIDDEN,
+        )
+
+    def test_get_work_request_private_workspace_authenticated(self):
+        """Work requests in private workspaces 200 to the authenticated."""
+        with context.disable_permission_checks():
+            private_workspace = self.playground.create_workspace(name="Private")
+        self.work_request.workspace = private_workspace
+        self.work_request.save()
+
+        self.client.force_login(self.playground.get_default_user())
+
+        response = self.client.get(
+            reverse("work-requests:detail", kwargs={"pk": self.work_request.id})
+        )
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+
     def test_get_work_request_no_retry_if_logged_out(self):
         """No retry link if the user is logged out."""
         self.work_request.task_name = "noop"
@@ -163,7 +198,7 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
         self.work_request.dynamic_task_data = {}
         self.work_request.save()
 
-        self.client.force_login(self.get_test_user())
+        self.client.force_login(self.playground.get_default_user())
         response = self.client.get(
             reverse("work-requests:detail", kwargs={"pk": self.work_request.id})
         )
@@ -178,7 +213,7 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
         self.work_request.status = WorkRequest.Statuses.ABORTED
         self.work_request.save()
 
-        self.client.force_login(self.get_test_user())
+        self.client.force_login(self.playground.get_default_user())
         response = self.client.get(
             reverse("work-requests:detail", kwargs={"pk": self.work_request.id})
         )
@@ -417,7 +452,9 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
             task_data={},
         )
         root = WorkRequest.objects.create_workflow(
-            template=template, data={}, created_by=self.get_test_user()
+            template=template,
+            data={},
+            created_by=self.playground.get_default_user(),
         )
         child_template = WorkflowTemplate.objects.create(
             name="child",
@@ -428,12 +465,12 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
         child = WorkRequest.objects.create_workflow(
             template=child_template,
             data={},
-            created_by=self.get_test_user(),
+            created_by=self.playground.get_default_user(),
             parent=root,
         )
         grandchildren = []
         for i in range(2):
-            wr = self.create_work_request(
+            wr = self.playground.create_work_request(
                 parent=child,
                 workflow_data_json={
                     "display_name": f"Lintian {i + 1}",
@@ -450,10 +487,13 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
 
-        root_link = f'<a href="/work-request/{root.id}/">noop</a>'
-        child_link = f'<a href="/work-request/{child.id}/">noop</a>'
+        def wr_url(pk: int) -> str:
+            return reverse("work-requests:detail", kwargs={"pk": pk})
+
+        root_link = f'<a href="{wr_url(root.id)}">noop</a>'
+        child_link = f'<a href="{wr_url(child.id)}">noop</a>'
         grandchild_links = [
-            f'<a href="/work-request/{grandchild.id}/">'
+            f'<a href="{wr_url(grandchild.id)}">'
             f'{grandchild.workflow_data.display_name}</a>'
             for grandchild in grandchildren
         ]
@@ -504,7 +544,9 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
             task_data={},
         )
         root = WorkRequest.objects.create_workflow(
-            template=template, data={}, created_by=self.get_test_user()
+            template=template,
+            data={},
+            created_by=self.playground.get_default_user(),
         )
         child_template = WorkflowTemplate.objects.create(
             name="child",
@@ -515,12 +557,12 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
         child = WorkRequest.objects.create_workflow(
             template=child_template,
             data={},
-            created_by=self.get_test_user(),
+            created_by=self.playground.get_default_user(),
             parent=root,
         )
         grandchildren = []
         for i in range(2):
-            wr = self.create_work_request(
+            wr = self.playground.create_work_request(
                 parent=child,
                 workflow_data_json={
                     "display_name": f"Lintian {i + 1}",
@@ -537,10 +579,13 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
 
         self.assertEqual(response.status_code, status.HTTP_200_OK)
 
-        root_link = f'<a href="/work-request/{root.id}/">noop</a>'
-        child_link = f'<a href="/work-request/{child.id}/">noop</a>'
+        def wr_url(pk: int) -> str:
+            return reverse("work-requests:detail", kwargs={"pk": pk})
+
+        root_link = f'<a href="{wr_url(root.id)}">noop</a>'
+        child_link = f'<a href="{wr_url(child.id)}">noop</a>'
         grandchild_links = [
-            f'<a href="/work-request/{grandchild.id}/">'
+            f'<a href="{wr_url(grandchild.id)}">'
             f'{grandchild.workflow_data.display_name}</a>'
             for grandchild in grandchildren
         ]
@@ -571,13 +616,15 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
 
     def test_workflow_manual_unblock_log(self) -> None:
         """A work request's manual unblock log is rendered if present."""
-        user = self.get_test_user()
-        self.add_user_permission(user, WorkRequest, "change_workrequest")
-        template = self.create_workflow_template("test", "noop")
+        user = self.playground.get_default_user()
+        self.playground.add_user_permission(
+            user, WorkRequest, "change_workrequest"
+        )
+        template = self.playground.create_workflow_template("test", "noop")
         root = WorkRequest.objects.create_workflow(
             template=template,
             data={},
-            created_by=self.get_test_user(),
+            created_by=self.playground.get_default_user(),
             status=WorkRequest.Statuses.RUNNING,
         )
         work_request = WorkRequest.objects.create_synchronization_point(
@@ -588,7 +635,7 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
         workflow_data.manual_unblock = WorkRequestManualUnblockData(
             log=[
                 WorkRequestManualUnblockLog(
-                    user_id=self.get_test_user().id,
+                    user_id=self.playground.get_default_user().id,
                     timestamp=timezone.now(),
                     notes="LGTM",
                     action=WorkRequestManualUnblockAction.ACCEPT,
@@ -608,7 +655,9 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
 
         tbody = tree.xpath("//table[@id='review-log']")[0].tbody[0]
         self.assertEqual(len(tbody.tr), 1)
-        self.assertEqual(tbody.tr.td[1], self.get_test_user().username)
+        self.assertEqual(
+            tbody.tr.td[1], self.playground.get_default_user().username
+        )
         self.assertHTMLEqual(
             html.tostring(tbody.tr.td[2], encoding="unicode"),
             '<td>'
@@ -630,6 +679,55 @@ class WorkRequestDetailViewTests(ViewTestMixin, TestCase):
             [tag.get("value") for tag in submit_buttons],
         )
 
+    def test_external_debsign_requires_signature(self) -> None:
+        """A blocked `ExternalDebsign` work request shows a prompt."""
+        user = self.playground.get_default_user()
+        work_request = self.playground.create_work_request(
+            task_type=TaskTypes.WAIT,
+            task_name="externaldebsign",
+            created_by=user,
+            status=WorkRequest.Statuses.BLOCKED,
+        )
+
+        self.client.force_login(user)
+        response = self.client.get(
+            reverse("work-requests:detail", kwargs={"pk": work_request.id})
+        )
+
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+        tree = self.assertHTMLValid(response)
+        requires_signature = tree.xpath("//div[@id='requires-signature']")[0]
+        self.assertTextContentEqual(
+            requires_signature.div[0], "Waiting for signature"
+        )
+        self.assertHTMLContentsEquivalent(
+            requires_signature.div[1],
+            f'<div class="card-body">'
+            f"Run <code>debusine provide-signature {work_request.id}</code> "
+            f"to sign this request."
+            f"</div>",
+        )
+
+    def test_external_debsign_wrong_user(self) -> None:
+        """A blocked `ExternalDebsign` only shows a prompt to its creator."""
+        work_request = self.playground.create_work_request(
+            task_type=TaskTypes.WAIT,
+            task_name="externaldebsign",
+            status=WorkRequest.Statuses.BLOCKED,
+        )
+        other_user = get_user_model().objects.create_user(
+            username="another", email="another@example.org"
+        )
+
+        self.client.force_login(other_user)
+        response = self.client.get(
+            reverse("work-requests:detail", kwargs={"pk": work_request.id})
+        )
+
+        self.assertEqual(response.status_code, status.HTTP_200_OK)
+        tree = self.assertHTMLValid(response)
+        self.assertEqual(tree.xpath("//div[@id='requires-signature']"), [])
+
 
 class WorkRequestListViewTests(ViewTestMixin, TestCase):
     """Tests for WorkRequestListView class."""
@@ -640,11 +738,14 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
     )
 
     @classmethod
+    @context.disable_permission_checks()
     def setUpTestData(cls):
         """Create workspaces used in the tests."""
         super().setUpTestData()
-        cls.private_workspace = cls.create_workspace(name="Private")
-        cls.public_workspace = cls.create_workspace(name="Public", public=True)
+        cls.private_workspace = cls.playground.create_workspace(name="Private")
+        cls.public_workspace = cls.playground.create_workspace(
+            name="Public", public=True
+        )
 
     def test_get_no_work_request(self):
         """No work requests in the server: 'No work requests' in response."""
@@ -655,12 +756,12 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
 
     def test_get_work_requests_all_authenticated_request(self):
         """Two work requests: all information appear in the response."""
-        public_work_request = self.create_work_request(
+        public_work_request = self.playground.create_work_request(
             task_name="noop",
             result=WorkRequest.Results.SUCCESS,
             workspace=self.public_workspace,
         )
-        private_work_request = self.create_work_request(
+        private_work_request = self.playground.create_work_request(
             task_name="noop", workspace=self.private_workspace
         )
 
@@ -676,7 +777,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
         # For all the workspaces does not display "Workspace: "
         self.assertNotContains(response, "Workspace: ")
 
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             response.context_data["object_list"],
             WorkRequest.objects.all().order_by("-created_at"),
         )
@@ -693,12 +794,12 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
 
     def test_get_work_requests_public_not_authenticated(self):
         """One work request: public one only."""
-        public_work_request = self.create_work_request(
+        public_work_request = self.playground.create_work_request(
             task_name="noop",
             result=WorkRequest.Results.SUCCESS,
             workspace=self.public_workspace,
         )
-        self.create_work_request(
+        self.playground.create_work_request(
             task_name="noop", workspace=self.private_workspace
         )
 
@@ -713,7 +814,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
 
     def test_get_work_requests_filter_by_bad_workspace(self):
         """Work request is created and filtered with an invalid name."""
-        self.create_work_request(task_name="noop")
+        self.playground.create_work_request(task_name="noop")
 
         response = self.client.get(
             reverse("work-requests:list") + "?workspace=does-not-exist"
@@ -722,7 +823,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
 
     def test_get_work_requests_filter_by_good_workspace(self):
         """Work request is created and filtered with a valid name."""
-        self.create_work_request(task_name="noop")
+        self.playground.create_work_request(task_name="noop")
 
         response = self.client.get(
             reverse("work-requests:list") + "?workspace=Public"
@@ -738,7 +839,9 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
             task_data={},
         )
         root = WorkRequest.objects.create_workflow(
-            template=template, data={}, created_by=self.get_test_user()
+            template=template,
+            data={},
+            created_by=self.playground.get_default_user(),
         )
         WorkRequest.objects.create_synchronization_point(
             parent=root, step="test"
@@ -748,7 +851,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
         self.assertEqual(response.status_code, status.HTTP_200_OK)
 
         tree = self.assertHTMLValid(response)
-        self.assertQuerysetEqual(response.context_data["object_list"], [root])
+        self.assertQuerySetEqual(response.context_data["object_list"], [root])
 
         table = tree.xpath("//table[@id='work_request-list']")[0]
         self.assertWorkRequestRow(table.tbody.tr[0], root)
@@ -756,7 +859,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
 
     def test_pagination(self):
         """Pagination is set up and rendered by the template."""
-        self.create_work_request(
+        self.playground.create_work_request(
             task_name="noop", workspace=self.public_workspace
         )
 
@@ -831,7 +934,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
                     reverse("work-requests:list")
                     + f"?status={work_request_status}"
                 )
-                self.assertQuerysetEqual(
+                self.assertQuerySetEqual(
                     response.context_data["work_request_list"], [work_request]
                 )
 
@@ -862,7 +965,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
             ],
         )
 
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             response.context_data["work_request_list"], [work_request]
         )
 
@@ -891,7 +994,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
             ],
         )
 
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             response.context_data["work_request_list"], [work_request]
         )
 
@@ -944,7 +1047,7 @@ class WorkRequestListViewTests(ViewTestMixin, TestCase):
             reverse("work-requests:list") + "?status=pending&arch=amd64"
         )
 
-        self.assertQuerysetEqual(
+        self.assertQuerySetEqual(
             response.context_data["work_request_list"],
             [work_request_amd64_2, work_request_amd64_1],
         )
@@ -1131,7 +1234,9 @@ class WorkRequestCreateViewTests(TestCase):
         except Exception as exc:
             error_message = f"Invalid task data: {exc}"
 
-        self.assertFormError(response, "form", "task_data", error_message)
+        self.assertFormError(
+            response.context["form"], "task_data", error_message
+        )
 
 
 class WorkRequestRetryViewTests(TestCase):
@@ -1145,9 +1250,9 @@ class WorkRequestRetryViewTests(TestCase):
     def setUpTestData(cls):
         """Initialize class data."""
         super().setUpTestData()
-        cls.user = cls.get_test_user()
+        cls.user = cls.playground.get_default_user()
         cls.workspace = Workspace.objects.earliest("id")
-        cls.work_request = cls.create_work_request(
+        cls.work_request = cls.playground.create_work_request(
             workspace=cls.workspace,
             task_name="noop",
             task_data={},
@@ -1249,8 +1354,8 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
     def setUpTestData(cls) -> None:
         """Initialize class data."""
         super().setUpTestData()
-        cls.user = cls.get_test_user()
-        template = cls.create_workflow_template("test", "noop")
+        cls.user = cls.playground.get_default_user()
+        template = cls.playground.create_workflow_template("test", "noop")
         cls.workflow = WorkRequest.objects.create_workflow(
             template=template,
             data={},
@@ -1305,7 +1410,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
         """Try unblocking a nonexistent work request."""
         pk = self.work_request.pk
         self.work_request.delete()
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
@@ -1322,7 +1429,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
 
     def test_invalid_form(self) -> None:
         """The view returns 400 if the form is invalid."""
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
@@ -1340,7 +1449,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
 
     def test_invalid_action(self) -> None:
         """The view returns 400 if the action parameter is invalid."""
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
@@ -1361,7 +1472,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
         """The view returns 400 if the work request cannot be unblocked."""
         self.work_request.unblock_strategy = WorkRequest.UnblockStrategy.DEPS
         self.work_request.save()
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
@@ -1381,7 +1494,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
 
     def test_accept(self) -> None:
         """Accept a blocked work request."""
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
@@ -1412,7 +1527,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
 
     def test_reject(self) -> None:
         """Reject a blocked work request."""
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
@@ -1443,7 +1560,9 @@ class WorkRequestUnblockViewTests(ViewTestMixin, TestCase):
 
     def test_record_notes_only(self) -> None:
         """Record notes on a blocked work request."""
-        self.add_user_permission(self.user, WorkRequest, "change_workrequest")
+        self.playground.add_user_permission(
+            self.user, WorkRequest, "change_workrequest"
+        )
         self.client.force_login(self.user)
 
         response = self.client.post(
