@@ -24,17 +24,11 @@ from unittest import mock
 from unittest.mock import MagicMock
 
 import aiohttp
+import tenacity
 from aiohttp import RequestInfo, WSMessage, WSMsgType
 from aiohttp.web import json_response
-from aiohttp.web_exceptions import (
-    HTTPForbidden,
-    HTTPInternalServerError,
-)
-
+from aiohttp.web_exceptions import HTTPForbidden, HTTPInternalServerError
 from multidict import CIMultiDict, CIMultiDictProxy
-
-import tenacity
-
 from yarl import URL
 
 import debusine.tasks.models as task_models
@@ -44,15 +38,22 @@ from debusine.client.models import (
     WorkRequestResponse,
     model_to_json_serializable_dict,
 )
-from debusine.tasks import BaseExternalTask, BaseTask, TaskConfigError
+from debusine.tasks import (
+    BaseExternalTask,
+    BaseTask,
+    Noop,
+    Sbuild,
+    TaskConfigError,
+)
 from debusine.tasks.models import TaskTypes, WorkerType
-from debusine.tasks.noop import Noop
-from debusine.tasks.sbuild import Sbuild
 from debusine.tasks.tests.helper_mixin import TestTaskMixin
 from debusine.test import TestCase
 from debusine.worker import Worker
 from debusine.worker.config import ConfigHandler
-from debusine.worker.system_information import system_metadata
+from debusine.worker.system_information import (
+    host_architecture,
+    system_metadata,
+)
 from debusine.worker.tests import server
 from debusine.worker.tests.server import ServerConfig
 
@@ -638,10 +639,15 @@ class WorkerTests(TestCase, server.DebusineAioHTTPTestCase):
         # The WorkRequest is executed...
         self.assertEqual(mocked_execute.call_count, 1)
 
-        # Check the workspace assigned in server.py
+        # Check the workspace and worker host architecture assigned to the
+        # task
         self.assertEqual(
             mocked_execute.call_args[0][0].workspace_name, "TestWorkspace"
         )
+        self.assertEqual(
+            mocked_execute.call_args[0][0].worker_host_architecture,
+            host_architecture(),
+        )
 
         # BaseTask.configure_server_access was called
         debusine_server = mocked_configure_server_access.mock_calls[0][1][1]
@@ -696,7 +702,8 @@ class WorkerTests(TestCase, server.DebusineAioHTTPTestCase):
         lock = threading.Lock()
         lock.acquire()
 
-        def wait(sbuild_self) -> None:  # noqa: U100
+        def wait(sbuild_self) -> None:
+            sbuild_self  # fake usage for vulture
             lock.acquire()
             lock.release()
 
